"""
Analysis and plotting tools for cerebellar cortex reconstructions.
"""

import matplotlib

from cerebellar_models.analysis.plots import Legend, Plot, ScaffoldPlot
from cerebellar_models.analysis.report import (
    LIST_CT_INFO,
    BSBReport,
    PlotTypeInfo,
    Report,
)
from cerebellar_models.analysis.spiking_results import (
    BasicSimulationReport,
    FiringRatesPlot,
    FrequencyPlot,
    ISIPlot,
    RasterPSTHPlot,
    SimResultsTable,
    SpikePlot,
    SpikeSimulationReport,
)
from cerebellar_models.analysis.structure_analysis import (
    CellPlacement3D,
    ConnectivityTable,
    PlacementTable,
    RunStructureReport,
    StructureReport,
    TablePlot,
)

matplotlib.use("Agg")

__all__ = [
    "Legend",
    "Plot",
    "ScaffoldPlot",
    "Report",
    "BSBReport",
    "PlotTypeInfo",
    "LIST_CT_INFO",
    "BasicSimulationReport",
    "FiringRatesPlot",
    "FrequencyPlot",
    "SpikePlot",
    "SpikeSimulationReport",
    "SimResultsTable",
    "ISIPlot",
    "RasterPSTHPlot",
    "TablePlot",
    "PlacementTable",
    "ConnectivityTable",
    "CellPlacement3D",
    "StructureReport",
    "RunStructureReport",
]



"""
Module for the configuration node of the Glomerulus to Granule ConnectionStrategy
"""

import itertools

import numpy as np
from bsb import (
    CfgReferenceError,
    ConfigurationError,
    ConnectionStrategy,
    ConnectivityError,
    InvertedRoI,
    config,
    pool_cache,
    refs,
)


class TooFewGlomeruliClusters(ConnectivityError):
    """
    Error raised when too few glomerulus clusters are available for a postsynaptic cell.
    """

    pass


@config.node
class ConnectomeGlomerulusGranule(InvertedRoI, ConnectionStrategy):
    """
    BSB Connection strategy to connect Glomerulus to Granule cells.
    With a convergence value set to `n`, this connection guarantees that each Granule cell connects
    to `n` unique Glomerulus clusters, where each Glomerulus cluster is connected to a different
    Mossy fiber.
    """

    radius = config.attr(type=int, required=True)
    """Radius of the sphere to filter the presynaptic chunks within it."""
    convergence: float = config.attr(type=float, required=True)
    """Convergence value between Glomeruli and Granule cells. 
        Corresponds to the mean number of Glomeruli that has a single Granule cell as target"""
    pre_glom_strats = config.reflist(refs.connectivity_ref, required=True)
    """Connection Strategies that links Pre-presyn cell to Glomeruli."""
    pre_cell_types = config.reflist(refs.cell_type_ref, required=True)
    """Celltype used for the pre-presyn cell."""
    depends_on: list[ConnectionStrategy] = config.reflist(refs.connectivity_ref)

    @config.property
    def depends_on(self):
        # Get the possibly missing `_depends_on` list.
        # Fixme: does not work with str.
        deps = getattr(self, "_depends_on", None) or []
        # Strat is required, but depends on a reference that isn't available when the config loads.
        strats = getattr(self, "pre_glom_strats", None)
        if strats is None:
            return deps
        else:
            return [*{*deps, *strats}]

    @depends_on.setter
    def depends_on(self, value):
        self._depends_on = value

    def _assert_dependencies(self):
        # assert dependency rule corresponds to mossy to glom
        for strat in self.pre_glom_strats:
            post_ct = strat.postsynaptic.cell_types
            if len(post_ct) != 1 or post_ct[0] not in self.presynaptic.cell_types:
                raise ConfigurationError(
                    "Postsynaptic cell of dependency rule does not match this rule's"
                    " presynaptic cell."
                )
        for strat in self.pre_glom_strats:
            pre_ct = strat.presynaptic.cell_types
            found = False
            for pre_pre_ct in self.pre_cell_types:
                if pre_pre_ct in pre_ct:
                    found = True
                    break
            if not found:
                raise ConfigurationError(
                    f"Presynaptic cells of dependency rule {strat} does not contain any of the provided presynaptic "
                    f"cell types: {[ct.name for ct in pre_ct]}."
                )

    def boot(self):
        self._assert_dependencies()

    def connect(self, pre, post):
        for post_ps in post.placement:
            self._connect_type(pre, post_ps)

    @pool_cache
    def load_connections(self):
        dict_cs = {}
        for pre_ct in self.presynaptic.cell_types:
            for strat in self.pre_glom_strats:
                for pre_pre_ct in self.pre_cell_types:
                    try:
                        cs = strat.get_output_names(pre_pre_ct, pre_ct)
                    except ValueError:
                        continue
                    if len(cs) != 1:
                        raise CfgReferenceError(
                            f"Only one connection set should be given from {strat.name} with type {pre_pre_ct.name}."
                        )
                    dict_cs[cs[0]] = list(
                        self.scaffold.get_connectivity_set(cs[0]).load_connections().all()
                    )
        return dict_cs

    def _get_pre_clusters(self, pre_ps):
        # Find the glomeruli clusters

        clusters = []
        unique_pres = []
        for strat in self.pre_glom_strats:
            for pre_ct in self.pre_cell_types:
                try:
                    cs = strat.get_output_names(pre_ct, pre_ps.cell_type)
                except ValueError:
                    continue
                if len(cs) != 1:
                    raise CfgReferenceError(
                        f"Only one connection set should be given from {strat.name} with type {pre_ct.name}."
                    )
                # find pre-glom connections where the postsyn chunk corresponds to the
                # glom-grc presyn chunk
                pre_locs, glom_locs = self.load_connections()[cs[0]]
                ct_uniques = np.unique(pre_locs[:, 0])
                unique_pres.extend(ct_uniques)

                for current in ct_uniques:
                    glom_idx = np.where(pre_locs[:, 0] == current)[0]
                    clusters.append(glom_locs[glom_idx, 0])

        return unique_pres, clusters

    def _connect_type(self, pre, post_ps):
        gran_pos = post_ps.load_positions()
        gran_morphos = post_ps.load_morphologies().iter_morphologies(cache=True, hard_cache=True)

        # Find the glomeruli clusters
        class presyn_dict:
            def __init__(cls, glom_pos, unique_pre, clusters):
                cls.glom_pos = glom_pos
                cls.unique_pre = unique_pre
                cls.clusters = clusters

        presyn_dicts = [
            presyn_dict(pre_ps.load_positions(), *self._get_pre_clusters(pre_ps))
            for pre_ps in pre.placement
        ]
        unique_pre = np.concatenate([np.arange(len(d.unique_pre)) for d in presyn_dicts])
        dict_ids = np.repeat(
            np.arange(len(presyn_dicts)), [len(d.unique_pre) for d in presyn_dicts]
        )
        if len(unique_pre) < self.convergence:
            raise TooFewGlomeruliClusters(
                "Less than 4 unique pre-presynaptic cells have been found. "
                "Check the densities of pre-presynaptic cells and glomeruli in "
                "the configuration file."
            )

        # TODO: implement random rounding and adapt tests.
        n_conn = int(np.round(len(gran_pos) * self.convergence))
        pre_locs = np.full((n_conn, 3), -1, dtype=int)
        post_locs = np.full((n_conn, 3), -1, dtype=int)
        selected_pre = np.full(n_conn, -1, dtype=int)
        ptr = 0
        for i, gr_pos, morpho in zip(itertools.count(), gran_pos, gran_morphos):
            # morpho should have enough dendrites to match convergence
            dendrites = morpho.get_branches()
            if len(dendrites) < self.convergence:
                raise ConnectivityError(
                    f"The postsynaptic morphology should have at least as many dendrites as the convergence value: {self.convergence}"
                )

            # Randomize the order of the clusters and dendrites
            cluster_idx = np.arange(0, len(unique_pre))
            np.random.shuffle(cluster_idx)
            dendrites_idx = np.arange(0, len(dendrites))
            np.random.shuffle(dendrites_idx)

            # The following loop connects a glomerulus from each cluster to a grc dendrite
            # until the convergence rule is reached.
            # First it checks for glomerulus that are close enough,
            # otherwise, it chooses the closest glomerulus from each remaining cluster.
            gr_connections = 0
            current_cluster = 0
            check_dist = True
            while gr_connections < self.convergence:
                if current_cluster >= len(cluster_idx):
                    # Not enough glom were found close enough to the GrC.
                    # Select from the remaining (more distant) gloms.
                    current_cluster = 0
                    check_dist = False
                nc = cluster_idx[current_cluster]
                loc_dict = presyn_dicts[dict_ids[nc]]
                clusters = loc_dict.clusters[unique_pre[nc]]
                dist = np.linalg.norm(gr_pos - loc_dict.glom_pos[clusters], axis=1)
                if check_dist:
                    # Try to select a cell from 4 clusters satisfying the conditions
                    close_indices = np.nonzero(dist < self.radius)[0]
                    if len(close_indices) == 0:
                        current_cluster += 1
                        continue
                    # Id of the glomerulus, randomly selected between the available ones
                    rnd = np.random.randint(low=0, high=len(close_indices))
                    id_glom = clusters[close_indices[rnd]]
                else:
                    # If there are some free dendrites, connect them to the closest glomeruli,
                    # even if they do not satisfy the geometric conditions.
                    # Id of the glomerulus, randomly selected between the available ones
                    id_glom = clusters[np.argmin(dist)]
                selected_pre[ptr + gr_connections] = dict_ids[nc]
                pre_locs[ptr + gr_connections, 0] = id_glom
                # Id of the granule cell
                post_locs[ptr + gr_connections, 0] = i
                # Select one of the 4 dendrites
                dendrite = dendrites[dendrites_idx[gr_connections]]
                post_locs[ptr + gr_connections, 1] = morpho.branches.index(dendrite)
                # Select the terminal point of the branch
                post_locs[ptr + gr_connections, 2] = len(dendrite) - 1

                gr_connections += 1
                # remove cluster used already
                cluster_idx = np.delete(cluster_idx, current_cluster)
            ptr += gr_connections
        for i, pre_ps in enumerate(pre.placement):
            self.connect_cells(
                pre_ps, post_ps, pre_locs[selected_pre == i], post_locs[selected_pre == i]
            )



"""
Module for the configuration node of the IO to purkinje cells AfterConnectivityHook
"""

import numpy as np
from bsb import AfterConnectivityHook, ConnectionStrategy, config, refs, types


@config.node
class DuplicateSynapses(AfterConnectivityHook):
    """
    BSB postprocessing to duplicate connections from a connection strategy into
    multiple synapses per pair.
    """

    conn_strategy: ConnectionStrategy = config.ref(refs.connectivity_ref, required=True)
    """Connection Strategy to on which to apply the postprocessing."""

    contacts = config.attr(type=types.distribution(), default=1)
    """Number or distribution determining the amount of synaptic contacts one cell will form on another"""

    def postprocess(self):
        for cs_name in self.conn_strategy.get_output_names():
            # Draw the number of connection to create
            n = int(self.contacts.draw(1)[0])
            cs = self.scaffold.get_connectivity_set(cs_name)
            pre_ps = cs.pre_type.get_placement_set()
            post_ps = cs.post_type.get_placement_set()
            if n > 1:
                pre_locs, post_locs = cs.load_connections().all()
                pre_locs = np.repeat(pre_locs, n - 1, axis=0)
                post_locs = np.repeat(post_locs, n - 1, axis=0)
                cs.connect(pre_ps, post_ps, pre_locs, post_locs)



"""
Module for the utility class for postsynaptically-sorted ConnectionStrategy
"""

import itertools

import numpy as np
from bsb import BoxTree, InvertedRoI, config


def get_close_chunks(chunk, target_chunks, radius):
    """
    Look for target chunks which are less than radius away from the current one.

    :param chunk: Source chunk.
    :type chunk: bsb.storage._chunks.Chunk
    :param target_chunks: Target chunk.
    :type target_chunks: Set[bsb.storage._chunks.Chunk]
    :radius: Maximum distance from the source chunk.
    :return: list of presynaptic chunks
    :rtype: list[bsb.storage._chunks.Chunk]
    """
    postsyn_chunk = [
        np.concatenate((chunk * chunk.dimensions - radius, (chunk + 1) * chunk.dimensions + radius))
    ]
    chunks_mbb = [np.concatenate((c, c + 1) * c.dimensions) for c in target_chunks]
    selected = np.array([i for i in BoxTree(chunks_mbb).query(postsyn_chunk)])
    return [c for i, c in enumerate(target_chunks) if i in selected]


class PresynDistStrat(InvertedRoI):
    """
    Mixin class that id used for ConnectionStrategy that deal with the connections for a pre- and
    post-synaptic pair sorting them by the post-synaptic cell chunk.
    """

    radius = config.attr(type=int, required=True)
    """Radius of the sphere to filter the presynaptic chunks within it."""

    def get_region_of_interest(self, chunk):
        """
        Finds all the presynaptic chunks that are within a sphere of defined radius, centered on the
        postsynaptic chunk.

        :param chunk: Postsynaptic chunk.
        :type chunk: bsb.storage._chunks.Chunk
        :return: list of presynaptic chunks
        :rtype: list[bsb.storage._chunks.Chunk]
        """

        chunks = set(
            itertools.chain.from_iterable(
                ct.get_placement_set().get_all_chunks() for ct in self.presynaptic.cell_types
            )
        )
        return get_close_chunks(chunk, chunks, self.radius)



"""
Module for the plots and reports related to the structural analysis of BSB scaffold.
"""

from typing import List, Tuple, Union

import numpy as np
from bsb import AfterConnectivityHook, CellType, ConnectivitySet, Scaffold, config, warn
from matplotlib import pyplot as plt

from cerebellar_models.analysis.plots import Legend, ScaffoldPlot
from cerebellar_models.analysis.report import LIST_CT_INFO, BSBReport, PlotTypeInfo


class TablePlot:
    """
    Mixin for plotting tables with matplotlib
    """

    _values = []
    """List of list of values from the table"""
    table_values = []
    """List of list of string values to put in the table"""
    rows = []
    """Names of the table's rows"""
    columns = []
    """Names of the table's columns"""

    def reset_table(self):
        """
        Update the values of the table.
        """
        self._values = []
        self.table_values = []
        self.rows = []

    def plot_table(self, **kwargs):
        """
        Plot the table in the Figure.
        """
        if len(self.table_values) == 0:
            warn("No values to plot", UserWarning)
            return
        dict_plot = dict(
            rowColours=np.full((len(self.rows), 3), [0.8, 0.8, 0.8]),
            rowLoc="right",
            colColours=np.full((len(self.columns), 3), [0.8, 0.8, 0.8]),
            loc="center",
        )
        dict_plot.update(kwargs)
        self.get_ax().table(
            cellText=self.table_values, rowLabels=self.rows, colLabels=self.columns, **dict_plot
        )


class PlacementTable(TablePlot, ScaffoldPlot):
    """
    Table plot of the results of the placement for BSB Scaffold.
    This includes the counts and density of each cell type.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold = None,
        dict_colors=None,
        dict_abv=None,
        **kwargs,
    ):
        super().__init__(
            fig_size,
            scaffold=scaffold,
            dict_colors=dict_colors,
            **kwargs,
        )
        self.columns = ["Cell counts", r"Cell densities [$\mu m^{-3}$]"]
        self.dict_abv = dict_abv or {}
        """Dictionary of abbreviations for cell types"""

    def extract_ct_name(self, ct: CellType):
        """
        Convert the name of a cell type to its abbreviation.
        """
        return self.dict_abv[ct.name] if ct.name in self.dict_abv else ct.name

    def update(self):
        super().update()
        self.reset_table()
        for i, ps in enumerate(self.scaffold.get_placement_sets()):
            ct = ps.cell_type
            volume = np.sum([p.volume() for place in ct.get_placement() for p in place.partitions])
            ct_name = self.extract_ct_name(ct)
            for labels in ps.get_unique_labels():
                count = ps.get_labelled(labels).size
                self.rows.append(self.get_labelled_ct_name(ct_name, labels))
                self._values.append([count, volume])
                self.table_values.append(
                    [
                        "{:.2E}".format(count),
                        "{:.2E}".format((count / volume) if volume > 0.0 else 0.0),
                    ]
                )

    def plot(self, **kwargs):
        super().plot()
        self.plot_table(**kwargs)

    def get_volumes(self):
        """
        Return a dictionary which gives for each cell type
        the volume occupied by its cells in the Scaffold.
        The plot needs to be updated.

        :rtype: Dict[str, int]
        """
        return {ct: line[1] for ct, line in zip(self.rows, self._values)}

    def get_counts(self):
        """
        Return a dictionary which gives for each cell type
        the number placed in the Scaffold.
        The plot needs to be updated.

        :rtype: Dict[str, int]
        """
        return {ct: line[0] for ct, line in zip(self.rows, self._values)}

    def get_densities(self):
        return {ct: line[0] / line[1] for ct, line in zip(self.rows, self._values)}


class ConnectivityTable(TablePlot, ScaffoldPlot):
    """
    Table plot of the results of the connectivity for BSB Scaffold.
    This includes for each pair of connected cell types:

    - the number of synapses formed
    - the number of synapses per unique pair of cell.
    - the convergence ratio defined as the mean number of afferent
      connections created with a single postsynaptic cell
    - the divergence ratio defined as the mean number of efferent
      connections created with a single presynaptic cell
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold = None,
        dict_colors=None,
        dict_abv=None,
        **kwargs,
    ):
        super().__init__(
            fig_size,
            scaffold=scaffold,
            dict_colors=dict_colors,
            **kwargs,
        )
        self.columns = ["Nb. Synapses", "Synapses per pair", "Convergence", "Divergence"]
        self.dict_abv = dict_abv or {}
        """Dictionary of abbreviations for cell types"""

    def extract_strat_name(self, ps: ConnectivitySet):
        """
        Convert the name of a connection into its abbreviation.
        Connection strategy name is assumed to be in the form A_B, where
        A is the presynaptic type name and B is the postsynaptic type name.
        """
        for cs in self.scaffold.configuration.connectivity:
            if cs in ps.tag:
                text = ps.tag.split(cs + "_", 1)[-1]
                if ps.tag == cs or "_" not in text:
                    splits = ps.tag
                else:
                    splits = text
                break
        splits = splits.split("_")
        tag = []
        to_convert = []

        def parse_remains():
            if len(to_convert) > 0:
                to_test = "_".join(to_convert) + "_cell"
                if to_test in self.dict_abv:
                    return [self.dict_abv[to_test]]
                else:
                    return to_convert
            return []

        for text in splits:
            if "to" == text:
                tag.extend(parse_remains())
                to_convert = []
                tag.append("to")
            else:
                to_convert.append(text)
                to_test = "_".join(to_convert)
                if to_test in self.dict_abv:
                    tag.append(self.dict_abv[to_test])
                    to_convert = []
        tag.extend(parse_remains())

        return " ".join(tag)

    def update(self):
        super().update()
        self.reset_table()
        for ps in self.scaffold.get_connectivity_sets():
            # Get the ConnectivityIterator for the current connectivity strategy
            cs = self.scaffold.get_connectivity_set(ps.tag).load_connections().as_globals()
            pre_locs, post_locs = cs.all()
            # Find the pairs of pre-post neurons (combos)
            # and count how many synapses there are between each pair (combo_counts)
            combos, combo_counts = np.unique(
                np.column_stack((pre_locs[:, 0], post_locs[:, 0])), axis=0, return_counts=True
            )

            # Find the unique post and pre neurons
            _, uniquePre_count = np.unique(combos[:, 0], axis=0, return_counts=True)
            _, uniquePost_count = np.unique(combos[:, 1], axis=0, return_counts=True)
            self.rows.append(self.extract_strat_name(ps))
            self._values.append([len(pre_locs), combo_counts, uniquePost_count, uniquePre_count])
            self.table_values.append(
                [
                    len(pre_locs),
                    r"{:.2} $\pm$ {:.2}".format(np.mean(combo_counts), np.std(combo_counts)),
                    r"{:.2} $\pm$ {:.2}".format(
                        np.mean(uniquePost_count), np.std(uniquePost_count)
                    ),
                    r"{:.2} $\pm$ {:.2}".format(np.mean(uniquePre_count), np.std(uniquePre_count)),
                ]
            )

    def plot(self, **kwargs):
        super().plot()
        self.plot_table(**kwargs)

    def get_synapse_counts(self):
        """
        Return a dictionary which gives for each connection name
        the number of synapses formed.
        The plot needs to be updated.

        :rtype: Dict[str, int]
        """
        return {ct: line[0] for ct, line in zip(self.rows, self._values)}

    def get_nb_synapse_per_pair(self):
        """
        Return a dictionary which gives for each connection name
        the number of synapses per unique pair of cell.
        The plot needs to be updated.

        :rtype: Dict[str, numpy.ndarray[int]]
        """
        return {ct: line[1] for ct, line in zip(self.rows, self._values)}

    def get_convergences(self):
        """
        Return a dictionary which gives for each connection name
        the divergence number of each unique postsynaptic cell.
        The plot needs to be updated.

        :rtype: Dict[str, numpy.ndarray[int]]
        """
        return {ct: line[2] for ct, line in zip(self.rows, self._values)}

    def get_divergences(self):
        """
        Return a dictionary which gives for each connection name
        the divergence number of each unique presynaptic cell.
        The plot needs to be updated.

        :rtype: Dict[str, numpy.ndarray[int]]
        """
        return {ct: line[3] for ct, line in zip(self.rows, self._values)}


class CellPlacement3D(ScaffoldPlot):
    """
    Plot the position of the cells in the scaffold in 3D space.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold = None,
        dict_colors=None,
        ignored_ct=None,
        **kwargs,
    ):
        super().__init__(fig_size, scaffold, dict_colors=dict_colors, **kwargs)
        self.ignored_ct = ignored_ct or ["mossy_fibers", "glomerulus", "ubc_glomerulus"]
        """List of cell type names to ignore in the plot."""

    def init_plot(self, **kwargs):
        if self.is_initialized:
            plt.close(self.figure)
        self.is_initialized = True
        self.is_plotted = False
        self.figure = plt.figure(figsize=self.fig_size, **kwargs)
        self.axes = self.figure.add_subplot(111, projection="3d")

    @staticmethod
    def set_axes_equal(ax):
        """
        Make axes of 3D plot have equal scale so that spheres appear as spheres,
        cubes as cubes, etc.

        :param ax: a matplotlib axis, e.g., as output from plt.gca().
        """

        x_limits = ax.get_xlim3d()
        y_limits = ax.get_ylim3d()
        z_limits = ax.get_zlim3d()

        x_range = abs(x_limits[1] - x_limits[0])
        x_middle = np.mean(x_limits)
        y_range = abs(y_limits[1] - y_limits[0])
        y_middle = np.mean(y_limits)
        z_range = abs(z_limits[1] - z_limits[0])
        z_middle = np.mean(z_limits)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5 * max([x_range, y_range, z_range])

        ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

    def plot(self, **kwargs):
        super().plot()
        ax = self.get_ax()
        grey = [0.6, 0.6, 0.6, 1.0]
        for i, ps in enumerate(self.scaffold.get_placement_sets()):
            ct = ps.cell_type
            if ct.name not in self.ignored_ct:
                positions = ps.load_positions()
                if len(positions) > 0:
                    u_labels = ps.get_unique_labels()
                    if len(u_labels) > 1:
                        color = [
                            self.labelled_dict_colors.get(
                                self.get_labelled_ct_name(ct.name, labels), grey
                            )
                            for labels in u_labels
                        ]
                        colors = np.ones((len(positions), 4))
                        for j, labels in enumerate(u_labels):
                            colors[ps.get_labelled(labels)] = color[j]
                        alphas = colors[:, -1]
                        colors = colors[:, :-1]
                    else:
                        if ct.name in self.dict_colors and len(self.dict_colors[ct.name]) == 3:
                            color = self.dict_colors[ct.name]
                            alpha = 1.0
                        else:
                            *color, alpha = self.dict_colors.get(ct.name, grey)
                        colors = np.repeat([color], len(positions), axis=0)
                        alphas = np.repeat([alpha], len(positions), axis=0)
                    scale = np.power(ct.spatial.radius, 2)
                    ax.scatter(
                        positions[:, 0],
                        positions[:, 1],
                        positions[:, 2],
                        c=colors,
                        alpha=alphas,
                        s=scale,
                    )
        ax.set_xlabel(r"x in $\mu m$")
        ax.set_ylabel(r"y in $\mu m$")
        ax.set_zlabel(r"z in $\mu m$")
        self.set_axes_equal(ax)
        ax.set_title("Placement results", fontsize=40)


class StructureReport(BSBReport):
    """
    Report of the scaffold neural network structure containing:

    - a table listing the number and density of each cell type placed.
    - a table listing the number, convergence and divergence of connections
      for each connected cell type pair
    - a 3D plot showing the soma location of each cell in the circuit.
    - a legend plot
    """

    def __init__(self, scaffold: Union[str, Scaffold], cell_type_info: List[PlotTypeInfo] = None):
        super().__init__(scaffold, cell_type_info)
        to_ignore = ["glomerulus", "ubc_glomerulus"]
        num_labelled_ct = sum(
            len(ps.get_unique_labels()) for ps in self.scaffold.get_placement_sets()
        )
        legend = Legend(
            (10, 0.6 * (num_labelled_ct - len(to_ignore)) / 3.0),
            3,
            dict_legend=dict(columnspacing=2.0, handletextpad=0.1, fontsize=20, loc="lower center"),
            dict_abbreviations=self.labelled_abbreviations,
        )
        density_table = PlacementTable(
            (5, 0.22 * (num_labelled_ct + 1)),
            scaffold=self.scaffold,
            dict_abv=self.abbreviations,
        )
        connectivity_table = ConnectivityTable(
            (10, 0.22 * (len(self.scaffold.get_connectivity_sets()) + 1)),
            scaffold=self.scaffold,
            dict_abv=self.abbreviations,
        )
        plot3d = CellPlacement3D((10, 10), scaffold=self.scaffold)
        self.add_plot("density_table", density_table)
        self.add_plot("connectivity_table", connectivity_table)
        self.add_plot("placement_3d", plot3d)
        self.add_plot("legend", legend)
        legend.dict_colors = plot3d.labelled_dict_colors.copy()
        legend.remove_ct(self.labelled_cell_names, to_ignore)

    def preprocessing(self):
        self.plots["legend"].set_axis_off()
        self.plots["density_table"].set_axis_off()
        self.plots["connectivity_table"].set_axis_off()


@config.node
class RunStructureReport(AfterConnectivityHook):
    """
    BSB postprocessing node to generate a scaffold structural report after running the connectivity jobs.
    """

    output_filename: str = config.attr(required=True)
    """Name of the pdf file to save the report."""

    def postprocess(self):
        report = StructureReport(self.scaffold, LIST_CT_INFO)
        report.print_report(self.output_filename)



"""
Module for abstract classes interfacing matplotlib plots.
"""

import abc
from typing import List, Tuple, Union

import numpy as np
from bsb import Scaffold
from matplotlib import patches
from matplotlib import pyplot as plt


class Plot(abc.ABC):
    """
    Matplotlib plot class interface.
    """

    def __init__(
        self,
        fig_size: Union[Tuple[float, float], np.ndarray],
        nb_rows: int = 1,
        nb_cols: int = 1,
        dict_colors=None,
        **kwargs,
    ):
        if nb_rows <= 0 or nb_cols <= 0:
            raise ValueError("Number of columns and rows must be strictly positive.")
        self.nb_rows = nb_rows
        """Number of sub-panel rows in the plot."""
        self.nb_cols = nb_cols
        """Number of sub-panel columns in the plot."""
        self.fig_size = fig_size
        """Tuple size of the figure in inches."""
        self.is_initialized = False
        """Flag to indicate if this plot figure has been initialized."""
        self.is_plotted = False
        """Flag to indicate if this plot figure has been plotted."""
        self.is_updated = False
        """Flag to indicate if this plot data has been updated."""
        self.figure = None
        """Matplotlib Figure of the plot."""
        self.axes = None
        """Matplotlib Axes of the plot"""
        self.dict_colors = {}
        """Dictionary of element name to their color"""
        if dict_colors is not None:
            for key, value in dict_colors.items():
                self.set_color(key, value)

    def clear(self):
        """
        Clear the figure axes
        """
        for ax in self.get_axes():
            ax.clear()
        self.is_plotted = False

    def init_plot(self, **kwargs):
        """
        Initialize the plot and return figure and axes.
        """
        if self.is_initialized:
            plt.close(self.figure)
        self.figure, self.axes = plt.subplots(
            nrows=self.nb_rows, ncols=self.nb_cols, figsize=self.fig_size, **kwargs
        )
        self.is_initialized = True

    def get_axes(self):
        """
        Return figure axes as a flat list.
        """
        if not self.is_initialized:
            self.init_plot()
        if self.axes is None:
            return []
        elif self.nb_cols == 1 and self.nb_rows == 1:
            return [self.axes]
        elif self.nb_cols == 1 or self.nb_rows == 1:
            return self.axes
        else:
            return [ax for col_ax in self.axes for ax in col_ax]

    def get_ax(self, idx: int = 0):
        """
        Return the axis at the given index.
        """
        if not self.is_initialized:
            self.init_plot()
        if idx >= self.nb_cols * self.nb_rows or idx < 0:
            raise IndexError(
                f"Index of matplotlib ax out of range. Max {self.nb_cols * self.nb_rows}"
            )
        if self.nb_cols == 1 and self.nb_rows == 1:
            return self.axes
        elif self.nb_cols == 1 or self.nb_rows == 1:
            return self.axes[idx]
        else:
            return self.axes[idx // self.nb_cols][idx % self.nb_cols]

    def set_color(self, key: str, color: np.ndarray[float]):
        """
        Set the color dictionary for a given key.
        Colors must be an array of type RGB or RGBA.
        """
        if len(color) != 3 and len(color) != 4:
            raise ValueError("Color must be an array of size 3 or 4.")
        self.dict_colors[key] = np.array(color, dtype=float)
        # colors are updated so the figure should be updated too.
        if self.is_plotted:
            self.clear()

    def save_figure(self, output_name: str, dpi: int, pad=0, **kwargs):
        """
        Save the figure as a file.
        """
        if not self.is_plotted:
            self.plot()
        self.figure.tight_layout(pad=pad)
        self.figure.savefig(output_name, dpi=dpi, facecolor="white", **kwargs)

    def update(self):
        """
        Update function to prepare the data before plotting.
        """
        self.is_updated = True

    def plot(self, *args, **kwargs):
        """
        Plot or replot the figure.
        Calls the update function if needed.
        """
        if not self.is_initialized:
            self.init_plot()
        if self.is_plotted:
            self.clear()
        if not self.is_updated:
            self.update()
        self.is_plotted = True

    def show(self):  # pragma:nocover
        """
        Show the figure.
        The figure will be plotted if needed.
        """
        if not self.is_plotted:
            self.plot()
        self.figure.show()

    def set_axis_off(self, axes=None):
        """
        Removes the borders of the provided axes.
        If none are provided, all axes' borders will be removed.

        :param axes: List of matplotlib axes
        """
        if axes is None:
            axes = self.get_axes()
        for ax in axes:
            ax.axis("off")


class ScaffoldPlot(Plot):
    """
    Matplotlib plot interface for BSB Scaffold.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        nb_rows: int = 1,
        nb_cols: int = 1,
        dict_colors: dict = None,
        **kwargs,
    ):
        super().__init__(fig_size, nb_rows, nb_cols, dict_colors, **kwargs)
        self.scaffold = scaffold

    def set_scaffold(self, scaffold):
        """
        Set the scaffold of the plot.
        """
        is_different = scaffold != self.scaffold
        if is_different:
            self.scaffold = scaffold
            self.is_updated = False
            if self.is_plotted:
                self.clear()
        return is_different

    @staticmethod
    def get_labelled_ct_name(ct_name, labels):
        extra = "_".join(labels)
        return ct_name + (extra if len(extra) == 0 else "_" + extra)

    @property
    def labelled_dict_colors(self):
        result = self.dict_colors.copy()
        for ps in self.scaffold.get_placement_sets():
            ct_name = ps.cell_type.name
            u_labels = ps.get_unique_labels()
            if len(u_labels) > 1:
                color = np.array(result[ct_name])
                del result[ct_name]
                for j, labels in enumerate(u_labels):
                    result[self.get_labelled_ct_name(ct_name, labels)] = color * (
                        [np.power(2 / 3, j)] * 3 + [1.0]
                    )
        return result


class Legend(Plot):
    """
    Patch Legend plot
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        nb_cols: int = 1,
        dict_colors: dict = None,
        dict_abbreviations: dict = None,
        dict_legend=None,
        **kwargs,
    ):
        super().__init__(fig_size, 1, 1, dict_colors, **kwargs)
        self.dict_legend = dict_legend or dict()
        self.dict_abbreviations = dict_abbreviations or dict()
        self.cols_legend = nb_cols

    def plot(self, **kwargs):
        super().plot()
        patchs = [
            patches.Patch(color=color[:3], label=label) for label, color in self.dict_colors.items()
        ]
        dict_plot = self.dict_legend.copy()
        dict_plot.update(kwargs)
        keys = [self.dict_abbreviations.get(k, k) for k in self.dict_colors.keys()]
        self.get_ax().legend(patchs, keys, ncol=self.cols_legend, **dict_plot)

    def remove_ct(self, to_keep: List[str], to_ignore: List[str] = None):
        """
        Remove cell types that are not in the list to keep or in the list to ignore

        :params to_keep List[str]: list of cell types to keep
        :params to_ignore List[str]: list of cell types to remove

        """
        to_ignore = to_ignore or []
        self.dict_colors = {
            ct: color
            for ct, color in self.dict_colors.items()
            if ct not in to_ignore and (ct in to_keep or ct + "_cell" in to_keep)
        }



"""
Module for the configuration node of the IO to molecular layer interneurons (MLI) ConnectionStrategy
"""

import numpy as np
from bsb import ConfigurationError, ConnectionStrategy, config, refs
from bsb.mixins import NotParallel


@config.node
class ConnectomeIO_MLI(NotParallel, ConnectionStrategy):
    """
    BSB Connection strategy to connect IO cells to molecular layer interneurons (MLI) cells through PC.
    IO cells which are connected to a PC should also connect to all the MLIs connected to this PC.
    """

    io_pc_connectivity = config.reflist(refs.connectivity_ref, required=True)
    """Connection Strategy that links IO to PC."""
    mli_pc_connectivity = config.reflist(refs.connectivity_ref, required=True)
    """List of Connection Strategies that links MLI to PC."""
    pre_cell_pc = config.ref(refs.cell_type_ref, required=True)
    """Celltype used for to represent PC."""
    depends_on: list[ConnectionStrategy] = config.reflist(refs.connectivity_ref)

    @config.property
    def depends_on(self):
        # Get the possibly missing `_depends_on` list.
        deps = getattr(self, "_depends_on", None) or []
        # Strat is required, but depends on a reference that isn't available when the config loads.
        strat_io = getattr(self, "io_pc_connectivity", None)
        strat_mli = getattr(self, "mli_pc_connectivity", None)
        return [*{*deps, *strat_io, *strat_mli}]

    @depends_on.setter
    def depends_on(self, value):
        self._depends_on = value

    def _assert_dependencies(self):
        # assert dependency rule corresponds to expected ones
        found_post = np.full(len(self.postsynaptic.cell_types), False)
        for strat in self.mli_pc_connectivity:
            post_ct = strat.postsynaptic.cell_types
            if len(post_ct) != 1 or post_ct[0] != self.pre_cell_pc:
                raise ConfigurationError(
                    f"PC cell type of the MLI to PC dependency rule does not correspond "
                    f"to the provided PC type for strat {strat.name}"
                )
            for i, post_ct in enumerate(self.postsynaptic.cell_types):
                if not found_post[i] and post_ct in strat.presynaptic.cell_types:
                    found_post[i] = True
        if not np.all(found_post):
            not_found = np.array([post_ct.name for post_ct in self.postsynaptic.cell_types])[
                ~found_post
            ]
            raise ConfigurationError(
                f"Postsynaptic cells: {not_found} "
                f"is not in any connection set of the MLI to PC dependency rules"
            )
        for strat in self.io_pc_connectivity:
            post_ct = strat.postsynaptic.cell_types
            if len(post_ct) != 1 or post_ct[0] != self.pre_cell_pc:
                raise ConfigurationError(
                    f"PC cell type of the IO to PC dependency rule does not correspond "
                    f"to the provided PC type for strategy {strat.name}"
                )
            for pre_ct in self.presynaptic.cell_types:
                if pre_ct not in strat.presynaptic.cell_types:
                    raise ConfigurationError(
                        f"Presynaptic cell: {pre_ct.name} is not in any connection set of "
                        f"the IO to PC dependency rule {strat.name}"
                    )

    def boot(self):
        self._assert_dependencies()

    def load_connectivity_set(self, connection_strat, cell_type):
        """
        Load the connection locations from a connection strategy that connects the provided cell type to PC.

        :param bsb.connectivity.strategy.ConnectionStrategy connection_strat: Connection strategy to load.
        :param str cell_type: Presynaptic cell type name.
        :return: A tuple containing:
                - an array of the presynaptic cell_type connection locations,
                - an array of the postsynaptic pc connection locations
        """
        cs = connection_strat.get_output_names(cell_type, self.pre_cell_pc)
        assert (
            len(cs) == 1
        ), f"Only one connection set should be given from {connection_strat.name}."
        cs = self.scaffold.get_connectivity_set(cs[0])
        return cs.load_connections().as_globals().all()

    def load_hemitype_connections(self, strategies, hemitype):
        """
        Load the connection locations for all the MLI to PC strategies.
        Will only keep one connection location information for each unique pair of MLI-PC.

        :param list[bsb.connectivity.strategy.ConnectionStrategy] strategies: Connection
         strategies to load.
        :param bsb.connectivity.strategy.HemitypeCollection hemitype: Hemitype
        :return: A tuple containing:
                - an array of the presynaptic connection locations
                - an array of the postsynaptic connection locations
                - an array of the placement set indexes
        """
        loc_all_pre = []
        loc_all_post = []
        ct_ps_ids = []
        # For each hemitype placement set
        for i, ps in enumerate(hemitype.placement):
            for strat in strategies:
                try:
                    # Fetch the corresponding connectivity set info if it exists
                    loc_pre, loc_post = self.load_connectivity_set(strat, ps.cell_type)
                except ValueError:
                    continue
                # keep one unique pair of mli to pc
                to_keep = np.unique(
                    np.concatenate([[loc_pre[:, 0]], [loc_post[:, 0]]]).T,
                    axis=0,
                    return_index=True,
                )[1]
                loc_all_pre.extend(loc_pre[to_keep])
                loc_all_post.extend(loc_post[to_keep])
                ct_ps_ids.extend(np.repeat(i, len(to_keep)))

        return np.asarray(loc_all_pre), np.asarray(loc_all_post), np.asarray(ct_ps_ids)

    def connect(self, pre, post):
        # We retrieve the connectivity data for the mli-pc connectivity and io-pc connectivity
        loc_io, loc_pc, ct_io_ids = self.load_hemitype_connections(self.io_pc_connectivity, pre)
        loc_mli, loc_mli_pc, ct_mli_ids = self.load_hemitype_connections(
            self.mli_pc_connectivity, post
        )

        for j, pre_ps in enumerate(pre.placement):
            loc_pc = loc_pc[:, 0][ct_io_ids == j]
            u_purkinje = np.unique(loc_pc)
            io_pc_list = []
            mli_per_pc_list = []
            grouped_ps_ids = []
            for current in u_purkinje:
                io_pc_list.append(loc_io[loc_pc == current][:, 0])
                mli_ids = loc_mli_pc[:, 0] == current
                mli_per_pc_list.append(loc_mli[mli_ids][:, 0])
                grouped_ps_ids.append(ct_mli_ids[mli_ids])

            max_len = np.max(
                [
                    [len(mli_ids), len(io_ids)]
                    for mli_ids, io_ids in zip(mli_per_pc_list, io_pc_list)
                ],
                axis=0,
            )
            max_connections = np.prod(max_len) * u_purkinje.size
            pre_locs = np.full((max_connections, 3), -1, dtype=int)
            post_locs = np.full((max_connections, 3), -1, dtype=int)
            ps_locs = np.full(max_connections, -1, dtype=int)
            ptr = 0
            for i, io_ids in enumerate(io_pc_list):
                ln = len(mli_per_pc_list[i])
                for current in io_ids:
                    pre_locs[ptr : ptr + ln, 0] = current
                    post_locs[ptr : ptr + ln, 0] = mli_per_pc_list[i]
                    ps_locs[ptr : ptr + ln] = grouped_ps_ids[i]
                    ptr = ptr + ln

            # because we are using global indices we need to extract the global ps
            for i, post_ps in enumerate(post.placement):
                post_ps = self.scaffold.get_placement_set(post_ps.cell_type)
                current = (ps_locs == i)[:ptr]
                self.connect_cells(
                    pre_ps, post_ps, pre_locs[:ptr][current], post_locs[:ptr][current]
                )



import os
from collections.abc import Mapping
from glob import glob
from os.path import basename, isdir, join, splitext

from bsb import parse_configuration_content_to_dict


def get_folders_in_folder(folder_path: str, excepts: set = None):
    """
    List all folders in a given folder path.

    :param str folder_path: folder path
    :param set[str] excepts: sub folders to ignore
    :return: list of folders
    :rtype: list[str]
    """
    if excepts is None:
        excepts = set()
    return [f for f in os.listdir(folder_path) if isdir(join(folder_path, f)) and f not in excepts]


def load_configs_in_folder(folder_path: str, recursive=True):
    """
    Load all BSB configuration as dictionaries in a given folder path.

    :param str folder_path: folder path
    :param bool recursive: if True will load also configurations in sub folders to ignore
    :return: dict linking each filename to its BSB configuration dictionary
    :rtype: dict
    """
    configs = {}

    for ext in ["yaml", "yml", "json"]:
        files = f"/**/*.{ext}" if recursive else f"/*.{ext}"
        configs.update(load_configs_from_files(glob(folder_path + files, recursive=True)))
    return configs


def load_configs_from_files(filenames):
    """
    Load all BSB configuration as dictionaries for a list of given filenames.

    :param list[str] filenames: list of configuration files
    :return: dict linking each filename to its BSB configuration dictionary
    :rtype: dict
    """
    configs = {}
    for filename in filenames:
        with open(filename, "r") as f:
            data = f.read()
            configs[splitext(basename(filename))[0]] = parse_configuration_content_to_dict(
                data, path=filename
            )[0]
    return configs


def deep_update(d: dict, u: Mapping):
    """
    Recursively update a dictionary d based on a dictionary u.
    u will overwrite any keys also in d.

    :return: merged dictionary
    :rtype: dict
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d



"""
Implementation of the BSB framework for cerebellar cortex reconstructions and simulations.
"""

import os

__version__ = "0.8.0"


def templates():  # pragma: nocover
    """
    :meta private:
    """
    return [os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))]


classmap = {
    "bsb.connectivity.strategy.ConnectionStrategy": {
        "mossy_glom": "cerebellar_models.connectome.to_glomerulus.ConnectomeMossyGlomerulus",
        "glom_gran": "cerebellar_models.connectome.glomerulus_granule.ConnectomeGlomerulusGranule",
        "golgi_glom": "cerebellar_models.connectome.golgi_glomerulus.ConnectomeGolgiGlomerulus",
        "glom_golgi": "cerebellar_models.connectome.glomerulus_golgi.ConnectomeGlomerulusGolgi",
        "ubc_glom": "cerebellar_models.connectome.to_glomerulus.ConnectomeUBCGlomerulus",
        "glom_ubc": "cerebellar_models.connectome.glomerulus_ubc.ConnectomeGlomerulusUBC",
        "io_mli": "cerebellar_models.connectome.io_molecular.ConnectomeIO_MLI",
    },
    "bsb.postprocessing.AfterPlacementHook": {
        "label_microzones": "cerebellar_models.placement.microzones.LabelMicrozones",
    },
    "bsb.postprocessing.AfterConnectivityHook": {
        "struct_report": "cerebellar_models.analysis.structure_analysis.RunStructureReport",
    },
}



"""
Module for the configuration node of the Golgi to Glomerulus ConnectionStrategy
"""

import itertools

import numpy as np
from bsb import (
    CfgReferenceError,
    ConfigurationError,
    ConnectionStrategy,
    ConnectivityError,
    config,
    pool_cache,
    refs,
)


@config.node
class ConnectomeGolgiGlomerulus(ConnectionStrategy):
    """
    BSB Connection strategy to connect Golgi cells to postsynaptic cells through Glomeruli.
    With a divergence value set to `n`, this connection guarantees that each golgi cell
    connects to all postsynaptic cells that are themselves connected to `n` unique Glomerulus.
    """

    divergence: float = config.attr(type=float, required=True)
    """Divergence value between Golgi cells and Glomeruli. 
        Corresponds to the mean number of Glomeruli targeted by a single Golgi cell"""
    radius: float = config.attr(type=float, required=True)
    """Radius of the sphere surrounding the Golgi cell soma in which glomeruli can be connected."""
    glom_post_strats = config.reflist(refs.connectivity_ref, required=True)
    """Connection Strategies that links Glomeruli to the postsynaptic cells."""
    glom_cell_types = config.reflist(refs.cell_type_ref, required=True)
    """Cell types used for the Glomeruli."""
    depends_on: list[ConnectionStrategy] = config.reflist(refs.connectivity_ref)

    @config.property
    def depends_on(self):
        # Get the possibly missing `_depends_on` list.
        deps = getattr(self, "_depends_on", None) or []
        # Strat is required, but depends on a reference that isn't available when the config loads.
        strat = getattr(self, "glom_post_strats", None)
        if strat is None:
            return deps
        else:
            return [*{*deps, *strat}]

    @depends_on.setter
    def depends_on(self, value):
        self._depends_on = value

    def connect(self, pre, post):
        for pre_ps in pre.placement:
            self._connect_type(pre_ps, post)

    def _assert_dependencies(self):
        # assert dependency rule corresponds to glom to post
        for glom_post_strat in self.glom_post_strats:
            # At least one glomerulus type should match for each strat.
            found = False
            for glom_cell_type in self.glom_cell_types:
                if found:
                    break
                for pre_ct in glom_post_strat.presynaptic.cell_types:
                    if pre_ct == glom_cell_type:
                        found = True
                        break
            if not found:
                raise ConfigurationError(
                    f"Presynaptic cell of dependency rule {glom_post_strat.name} should match "
                    f"at least one of the provided glom_cell_types: "
                    f"{[glom_cell_type.name for glom_cell_type in self.glom_cell_types]}."
                )

            post_ct = glom_post_strat.postsynaptic.cell_types
            found = False
            for ct in self.postsynaptic.cell_types:
                if ct in post_ct:
                    found = True
                    break
            if not found:
                raise ConfigurationError(
                    f"The dependency rule {glom_post_strat.name} does not connect glomeruli to this connection's "
                    f"postsynaptic cell: {post_ct.name}."
                )

    def boot(self):
        self._assert_dependencies()

    @pool_cache
    def load_connections(self):
        dict_cs = {}
        for grc_ct in self.postsynaptic.cell_types:
            for strat in self.glom_post_strats:
                for inter_ct in self.glom_cell_types:
                    try:
                        cs = strat.get_output_names(inter_ct, grc_ct)
                    except ValueError:
                        continue
                    if len(cs) != 1:
                        raise CfgReferenceError(
                            f"Only one connection set should be given from {strat.name} with type {inter_ct.name}."
                        )
                    dict_cs[cs[0]] = list(
                        self.scaffold.get_connectivity_set(cs[0]).load_connections().all()
                    )
        return dict_cs

    def _combine_unique_gloms(self, glom_ids, loc_glom_pos, loc_post_locs):
        # Extract unique glomerulus groups
        unique_gloms = np.unique(glom_ids)
        loc_post_locs = np.asarray(loc_post_locs, dtype=object)
        post_locs = []
        glom_pos = np.zeros((len(unique_gloms), 3), dtype=float)
        for c, u_glom in enumerate(unique_gloms):
            ids = np.where(glom_ids == u_glom)[0]
            post_locs.append(
                [locs for locs in np.asarray(np.concatenate(loc_post_locs[ids]), dtype=int)]
            )
            glom_pos[c] = loc_glom_pos[ids[0]]

        return (unique_gloms, glom_pos, post_locs)

    def _get_glom_cluster(self, chunk, post_ps, glom_type):
        # Get the glom_to_post connections
        glom_ids = []
        post_locs = []
        loc_glom_pos = np.empty((0, 3))
        for glom_post_strat in self.glom_post_strats:
            try:
                # Extract connection sets that link glom_type to the postsynaptic cell type.
                cs = glom_post_strat.get_output_names(glom_type, post_ps.cell_type)
            except ValueError:
                continue
            if len(cs) != 1:
                raise CfgReferenceError(
                    f"Only one connection set should be given from {glom_post_strat.name}."
                )

            # We filter glomeruli in chunks which are less than radius away from the current one.
            loc_glom_locs, loc_post_locs = self.load_connections()[cs[0]]

            # Filter unique glomeruli connections and grouping postsynaptic targets
            loc_glom_locs, ids = np.unique(loc_glom_locs[:, 0], return_inverse=True)
            glom_ids.extend(loc_glom_locs)
            # Sorting connections by presynaptic glomeruli id
            sorting = np.argsort(ids)
            # Returns the index positions of each glomerulus group in the sorted connection array
            _, ids = np.unique(ids[sorting], return_index=True)
            # Sorts the postsynaptic info and then splitting it according to the presynaptic ids
            post_locs.extend(np.split(loc_post_locs[sorting], ids[1:]))

            # Since we use global ids there is no need to filter chunks of the presynaptic cell placement set
            loc_glom_pos = np.concatenate(
                [
                    self.scaffold.get_connectivity_set(cs[0])
                    .pre_type.get_placement_set()
                    .load_positions()[loc_glom_locs],
                    loc_glom_pos,
                ]
            )

        return self._combine_unique_gloms(glom_ids, loc_glom_pos, post_locs)

    def _connect_type(self, pre_ps, post):
        # Chunks are sorted presynaptically so there should be only one chunk
        chunk = pre_ps.get_loaded_chunks()
        if len(chunk) != 1:
            ConnectivityError("There should be exactly one presynaptic chunk")
        chunk = chunk[0]

        # Extract all unique glomeruli connections from connection strategy dependencies grouped by glomeruli
        glom_pos = np.empty([0, 3])
        postsyn_locs = []
        ps_ids = []
        for glom_cell_type in self.glom_cell_types:
            ct_glom_ids = []
            ct_glom_pos = np.empty([0, 3])
            ct_postsyn_locs = []
            ct_ps_ids = []
            for i, post_ps in enumerate(post.placement):
                (loc_glom_ids, loc_glom_pos, loc_post_locs) = self._get_glom_cluster(
                    chunk, post_ps, glom_cell_type
                )
                ct_glom_ids.extend(loc_glom_ids)
                ct_glom_pos = np.concatenate([loc_glom_pos, ct_glom_pos])
                ct_postsyn_locs.extend(loc_post_locs)
                ct_ps_ids.extend([np.repeat(i, len(ids)) for ids in loc_post_locs])
                i += 1
            # group gloms from different postsynaptic cell type together (e.g. UBC, GrC)
            u_glom_ids, ct_glom_pos, ct_postsyn_locs = self._combine_unique_gloms(
                ct_glom_ids, ct_glom_pos, ct_postsyn_locs
            )
            glom_pos = np.concatenate([ct_glom_pos, glom_pos])
            postsyn_locs.extend(ct_postsyn_locs)
            ct_ps_ids = np.asarray(ct_ps_ids, dtype=object)
            for u_glom in u_glom_ids:
                ids = np.where(ct_glom_ids == u_glom)[0]
                ps_ids.append(np.concatenate(ct_ps_ids[ids].tolist(), dtype=int))

        golgi_pos = pre_ps.load_positions()
        # Cache morphologies and generate the morphologies iterator
        golgi_morphos = pre_ps.load_morphologies().iter_morphologies(cache=True, hard_cache=True)

        # TODO: implement random rounding and adapt tests.
        num_glom_to_connect = np.min([int(self.divergence), len(postsyn_locs)])
        if num_glom_to_connect == 0:
            raise ConnectivityError(
                "The resolved potential targets or the divergence value should be greater than 0."
            )
        n_conn = (
            len(golgi_pos)
            * num_glom_to_connect
            * np.max([len(post_conn) for post_conn in postsyn_locs])
        )
        pre_locs = np.full((n_conn, 3), -1, dtype=int)
        post_locs = np.full((n_conn, 3), -1, dtype=int)
        selected_ps = np.full(n_conn, -1, dtype=int)
        ptr = 0
        for i, golgi, morpho in zip(itertools.count(), golgi_pos, golgi_morphos):
            # Find terminal branches
            axon_branches = morpho.get_branches()
            terminal_branches_ids = np.nonzero([b.is_terminal for b in axon_branches])[0]
            axon_branches = np.take(axon_branches, terminal_branches_ids, axis=0)
            terminal_branches_ids = np.array([morpho.branches.index(b) for b in axon_branches])

            # Find the point-on-branch ids of the tips
            tips_coordinates = np.array([len(b.points) - 1 for b in axon_branches])

            # Compute and sort the distances between the golgi soma and the glomeruli
            to_connect = np.linalg.norm(golgi - glom_pos, axis=1)
            sorting = np.argsort(to_connect)
            to_connect = sorting[to_connect[sorting] <= self.radius]
            # Keep the closest glomeruli
            to_connect = to_connect[:num_glom_to_connect]
            # For each glomerulus, connect the corresponding postsyn cells directly to the current Golgi
            for post_conn in to_connect:
                take_post = np.array(postsyn_locs[post_conn])
                post_to_connect = len(take_post)
                # Select postsyn cells ids
                post_locs[ptr : ptr + post_to_connect] = take_post
                selected_ps[ptr : ptr + post_to_connect] = ps_ids[post_conn]
                # Select Golgi axon branch
                pre_locs[ptr : ptr + post_to_connect, 0] = i
                ids_branches = np.random.randint(
                    low=0, high=len(axon_branches), size=post_to_connect
                )
                pre_locs[ptr : ptr + post_to_connect, 1] = terminal_branches_ids[ids_branches]
                pre_locs[ptr : ptr + post_to_connect, 2] = tips_coordinates[ids_branches]
                ptr += post_to_connect

        for i, ps in enumerate(post.placement):
            selected = selected_ps[:ptr] == i
            # We use the scaffold placement set for each post cell type so that the global postsynaptic ids are used
            self.connect_cells(
                pre_ps,
                self.scaffold.get_placement_set(ps.cell_type),
                pre_locs[:ptr][selected],
                post_locs[:ptr][selected],
            )



import copy
import os
from enum import Enum
from os.path import abspath, dirname, join

import click
import numpy as np
from bsb import (
    Configuration,
    ConfigurationError,
    format_configuration_content,
    parse_configuration_file,
)

from cerebellar_models import __version__
from cerebellar_models.utils import (
    deep_update,
    get_folders_in_folder,
    load_configs_in_folder,
)

ROOT_FOLDER = dirname(dirname(abspath(__file__)))
CONFIGURATION_FOLDER = join(ROOT_FOLDER, "configurations")


class TypeTermElem(Enum):
    """
    Enum for all supported survey widgets
    """

    Selection = 1
    Basket = 2
    Text = 3
    Number = 4
    Boolean = 5


class CerebOption:
    """
    Class to store the information of an element of a survey form
    """

    def __init__(
        self,
        name: str,
        title: str,
        choices=None,
        default_value=None,
        type_term: TypeTermElem = TypeTermElem.Selection,
    ):
        self.name = name
        """Name of the option"""
        self.title = title
        """Help text for the option"""
        if type_term is TypeTermElem.Number:
            choices = [0]
        elif type_term is TypeTermElem.Text:
            choices = [""]
        elif type_term is TypeTermElem.Boolean:
            choices = [False, True]
        elif choices is None:
            raise TypeError("Provide a list of choices for Selection or Basket options")
        self.choices = np.array(choices)
        """List of possible choices for this option"""
        self.value = default_value or (self.choices[0] if type_term != TypeTermElem.Basket else [])
        """Current value for the option"""
        self.type = type_term
        """Type of the option"""

    def get_widget(self):  # pragma: no cover
        """
        Return the survey widget for this option
        """
        import survey

        if self.type == TypeTermElem.Basket:
            return survey.widgets.Basket(
                options=self.choices, active=np.where(np.isin(self.choices, self.value))[0]
            )
        elif self.type == TypeTermElem.Selection:
            return survey.widgets.Select(options=self.choices)
        elif self.type == TypeTermElem.Text:
            return survey.widgets.Input(value=str(self.value))
        elif self.type == TypeTermElem.Boolean:
            return survey.widgets.Inquire(default=bool(self.value))
        elif self.type == TypeTermElem.Number:
            return survey.widgets.Numeric(value=float(self.value), decimal=True)


class MicrozonesParams:
    """Class to save the parameters for the micro-zones labelling"""

    def __init__(
        self,
        labels: list[str] = None,
        cell_types: list[str] = None,
    ):
        self.labels = labels or []
        """Labels to associate to each microzone"""
        self.cell_types = cell_types or []
        """Cell types to split into microzones"""
        self.conn_duplicated = {}
        """Duplicated connections obtained from the cell type labelling"""


def print_panel(options, title="Configure your cerebellum circuit."):  # pragma: no cover
    """
    Print a survey form based on a list of options.
    The result of the form will be saved in its options.

    :param list[CerebOption] options: List of options to display
    :param str title: Title to display on top of the form
    """
    import survey

    form = survey.routines.form(
        title, form={option.name: option.get_widget() for option in options}
    )
    for option in options:
        if option.type == TypeTermElem.Basket:
            option.value = option.choices[np.array(list(form[option.name]), dtype=int)]
        elif option.type == TypeTermElem.Selection:
            option.value = option.choices[form[option.name]]
        else:
            option.value = form[option.name]


@click.group(help="Cerebellar models CLI")
@click.version_option(__version__)
def app():  # pragma: no cover
    """The main CLI entry point"""
    pass


EXISTING_DIR_PATH = click.Path(
    exists=True, readable=True, writable=True, file_okay=False, resolve_path=True
)
AVAILABLE_SPECIES = click.Choice(get_folders_in_folder(CONFIGURATION_FOLDER), case_sensitive=True)
AVAILABLE_EXTENSIONS = click.Choice(["yaml", "json"], case_sensitive=True)


def _configure_species(species):
    main_options = [
        CerebOption(
            "Species",
            "Select a species from the following list:",
            AVAILABLE_SPECIES.choices,
            species,
        ),
    ]
    print_panel(main_options, "Select your configuration's species")
    return main_options[0].value


def _configure_cell_types(species_folder, config_cell_types, add_microzones: bool):
    cell_type_names = []
    for filename1, config_1 in config_cell_types.items():
        cell_types1 = list(config_1["cell_types"].keys())
        for filename2, config_2 in config_cell_types.items():
            if filename1 != filename2 and np.all(
                np.isin(cell_types1, list(config_2["cell_types"].keys()))
            ):
                cell_type_names.insert(0, filename1)
                break
        if filename1 not in cell_type_names:
            cell_type_names.append(filename1)

    species_options = [
        CerebOption(
            "State",
            "Select the state of the subject to model from the following list:",
            get_folders_in_folder(species_folder, {"cell_types"}),
            default_value="awake",
        ),
        CerebOption(
            "Extra cell types",
            "Select the optional cell types that you want in the final configuration from the following list:",
            cell_type_names,
            type_term=TypeTermElem.Basket,
            default_value=["dcn", "io"],
        ),
        CerebOption(
            "Add microzones?",
            "Split your circuit into microzones",
            type_term=TypeTermElem.Boolean,
            default_value=add_microzones,
        ),
    ]
    print_panel(
        species_options, "Select the state of the subject and the cell types to add in the circuit"
    )
    return [option.value for option in species_options]


def _update_cell_types(configuration, cell_types, config_cell_types):
    for cell_type in cell_types:
        config_ = config_cell_types[cell_type]
        for k, v in config_.items():  # update within the main components
            if k == "network":
                for net_key, net_v in v.items():
                    if net_key in ["x", "y", "z"]:
                        configuration[k][net_key] = max(configuration[k][net_key], net_v)
            else:
                if k not in configuration:
                    configuration[k] = v
                else:
                    deep_update(configuration[k], v)
    return configuration


def _add_microzones(configuration, micro_params: MicrozonesParams):
    # Keep only the labelled cells that are in the final config.
    cells_found = []
    for cell_type in configuration["cell_types"]:
        if cell_type in micro_params.cell_types:
            cells_found.append(cell_type)
    micro_params.cell_types = cells_found

    # add labelling strat
    configuration["after_placement"] = {
        "label_microzones": {
            "strategy": "cerebellar_models.placement.microzones.LabelMicrozones",
            "cell_types": cells_found,
            "labels": micro_params.labels,
        }
    }

    # The connectivity rules that have both their pre- and post- synaptic cell types
    # labelled are updated to filter the first label (no change of name)
    first_label = micro_params.labels[0]
    for strat_name, strat in configuration["connectivity"].items():
        found = True
        for hemitype in ["presynaptic", "postsynaptic"]:
            if not found:
                break
            for cell_type in strat[hemitype]["cell_types"]:
                if cell_type not in micro_params.cell_types:
                    found = False
                    break
        if found:
            duplicate_rules = [copy.deepcopy(strat)] * len(micro_params.labels[1:])
            for hemitype in ["presynaptic", "postsynaptic"]:
                if "labels" not in strat[hemitype]:
                    strat[hemitype]["labels"] = []
                    for duplicate_rule in duplicate_rules:
                        duplicate_rule[hemitype]["labels"] = []

                strat[hemitype]["labels"].append(first_label)
                for label, duplicate_rule in zip(micro_params.labels[1:], duplicate_rules):
                    duplicate_rule[hemitype]["labels"].append(label)
            micro_params.conn_duplicated[strat_name] = duplicate_rules

    # Duplicate the connectivity rule for each extra label
    for strat_name, duplicate_rules in micro_params.conn_duplicated.items():
        for i in range(len(duplicate_rules)):
            new_name = f"{strat_name}_{micro_params.labels[1 + i]}"
            configuration["connectivity"][new_name] = duplicate_rules[i]
            micro_params.conn_duplicated[strat_name][i] = new_name  # keep only new names

    # Update references to the connectivity rules
    for strat_name, strat in configuration["connectivity"].items():
        for k, v in strat.items():
            if isinstance(v, list):
                for elem in v:
                    if elem in micro_params.conn_duplicated:
                        new_names = [f"{elem}_{label}" for label in micro_params.labels[1:]]
                        strat[k].extend(new_names)
    return configuration


def _configure_simulations(config_simulations):
    simulation_names = list(
        set(
            [
                f"{simulator}_{simu}"
                for simulator, v in config_simulations.items()
                for file_ in v["simulations"].values()
                for simu in file_["simulations"].keys()
            ]
        )
    )
    simulator_options = [
        CerebOption(
            "Simulations",
            "Select the simulations(s) that you want to perform from the following list:",
            simulation_names,
            type_term=TypeTermElem.Basket,
            default_value=["nest_basal_activity"],
        ),
    ]
    print_panel(
        simulator_options, "Select the simulations(s) that you want your circuit to perform"
    )
    return simulator_options[0].value


def _configure_sim_params(config_simulations, simulation_names, micro_params: MicrozonesParams):
    dict_sim = {"simulations": {}}
    choices = {}
    for sim_name in simulation_names:
        sim_name = str(sim_name)
        simulator, simulation = sim_name.split("_", 1)
        for k, v in config_simulations[simulator]["simulations"].items():
            if simulation in v["simulations"]:
                for sim, params in v.items():
                    if sim == "simulations":
                        dict_sim[sim][simulation] = params[simulation]
                    else:
                        dict_sim[sim] = params
        simulation_options = [
            CerebOption(
                "Cell models",
                f"Select the model of neuron for the simulation {sim_name} from the following list:",
                list(config_simulations[simulator]["cell_models"].keys()),
                # default_value="eglif_cond_alpha_multisyn"
            ),
            CerebOption(
                "Connection models",
                f"Select the model of synapse for the simulation {sim_name} from the following list:",
                list(config_simulations[simulator]["connection_models"].keys()),
                default_value="tsodyks2_synapse",
            ),
        ]
        print_panel(
            simulation_options,
            f"Select the neuron and synapse model to use during the simulation {sim_name}.",
        )
        deep_update(
            dict_sim, config_simulations[simulator]["cell_models"][simulation_options[0].value]
        )
        deep_update(
            dict_sim,
            config_simulations[simulator]["connection_models"][simulation_options[1].value],
        )
        choices[sim_name] = [c.value for c in simulation_options]
        # Add simulator to simulation name so that we avoid duplicates
        dict_sim["simulations"][sim_name] = dict_sim["simulations"][simulation]
        del dict_sim["simulations"][simulation]

        # duplicate connections linked to micro-zones
        simulation = dict_sim["simulations"][sim_name]
        for conn, new_names in micro_params.conn_duplicated.items():
            for new_name in new_names:
                simulation["connection_models"][new_name] = copy.deepcopy(
                    simulation["connection_models"][conn]
                )
        # duplicate devices to record labelled cells separately
        names = list(simulation["devices"].keys())
        for cell_type in micro_params.cell_types:
            for name in names:
                device = simulation["devices"][name]
                if (
                    "_record" in name
                    and device["targetting"]["strategy"] == "cell_model"
                    and cell_type in device["targetting"]["cell_models"]
                ):
                    device["targetting"]["strategy"] = "by_label"
                    device["targetting"]["labels"] = [micro_params.labels[0]]
                    for label in micro_params.labels[1:]:
                        new_name = f"{name.split('_record')[0]}_{label}_record"
                        simulation["devices"][new_name] = copy.deepcopy(device)
                        simulation["devices"][new_name]["targetting"]["labels"] = [label]
                    break

    return dict_sim, choices


def _clear_unnecessary_params(configuration):
    dict_cells = {}
    dict_conns = {}
    dict_devices = {}
    for sim_name, simulation in configuration["simulations"].items():
        dict_cells[sim_name] = []
        dict_conns[sim_name] = []
        dict_devices[sim_name] = []
        for cell in simulation["cell_models"]:
            if cell not in configuration["cell_types"]:
                dict_cells[sim_name].append(cell)
        for syn in simulation["connection_models"]:
            found = False
            for strat in configuration["connectivity"]:
                if strat in syn:
                    loc_strat = configuration["connectivity"][strat]
                    simple_conn = (
                        len(loc_strat["presynaptic"]["cell_types"]) == 1
                        and len(loc_strat["postsynaptic"]["cell_types"]) == 1
                    )
                    if simple_conn and strat == syn:
                        found = True
                        break
                    elif simple_conn != (strat == syn):
                        continue
                    cells = syn.split(strat + "_", 1)[1].split("_to_")
                    if (
                        cells[0] in loc_strat["presynaptic"]["cell_types"]
                        and cells[1] in loc_strat["postsynaptic"]["cell_types"]
                    ):
                        found = True
                        break
            if not found:
                dict_conns[sim_name].append(syn)
        for device_name, device in simulation["devices"].items():
            for target in device["targetting"]["cell_models"]:
                if target not in configuration["cell_types"]:
                    dict_devices[sim_name].append(device_name)

    for sim_name, to_remove in dict_cells.items():
        for cell in to_remove:
            del configuration["simulations"][sim_name]["cell_models"][cell]
        for syn in dict_conns[sim_name]:
            del configuration["simulations"][sim_name]["connection_models"][syn]
        for device in dict_devices[sim_name]:
            del configuration["simulations"][sim_name]["devices"][device]

    return configuration


def _write_config(configuration, output_folder, extension):
    output_options = []
    if output_folder is None:
        output_options.append(
            CerebOption(
                "Configuration folder",
                "Configure the folder in which to put the configuration file",
                default_value=os.getcwd(),
                type_term=TypeTermElem.Text,
            )
        )
    else:
        output_folder = abspath(output_folder)
        print(f"Output folder chosen: {output_folder}")
    if extension is None:
        output_options.append(
            CerebOption(
                "File extension",
                "Select an extension from the following list:",
                AVAILABLE_EXTENSIONS.choices,
                default_value=AVAILABLE_EXTENSIONS.choices[0],
            )
        )
    else:
        print(f"File extension chosen: {extension}")
    if len(output_options) > 0:
        print_panel(
            output_options,
            "Configure the folder in which to put the configuration file and its extension.",
        )
        output_folder = output_folder or output_options[0].value
        extension = extension or output_options[-1].value
    filename = os.path.join(output_folder, f"circuit.{extension}")
    try:
        configuration = Configuration.default(**configuration)  # Check that the configuration works
    except ConfigurationError as e:  # pragma: no cover
        raise ValueError("A BSB error happened when loading your cerebellar circuit") from e
    with open(filename, "w") as outfile:
        outfile.write(format_configuration_content(extension, configuration))
    print(f"Created the BSB configuration file: {filename}")


@app.command(help="Create a BSB configuration file for your cerebellum circuit.")
@click.option(
    "--species",
    type=AVAILABLE_SPECIES,
    required=False,
    help="Species to reconstruct the circuit from.",
)
@click.option(
    "--output_folder",
    type=EXISTING_DIR_PATH,
    required=False,
    help="Path where to write the output configuration file.",
)
@click.option(
    "--extension",
    type=AVAILABLE_EXTENSIONS,
    required=False,
    help="Extension for the configuration file.",
)
@click.option(
    "--microzones",
    required=False,
    is_flag=True,
    help="Split your circuit into 2 separated microzones.",
)
def configure(
    species: str = None,
    output_folder: str = None,
    extension: str = None,
    microzones: bool = False,
):
    """
    Resolve a canonical cerebellum configuration file for BSB based on user choices.

    :param str species: species to reconstruct the circuit from.
    :param str output_folder: path where to write the output configuration file.
    :param str extension: extension for the configuration file.
    :param bool microzones: whether to split the circuit into microzones.
    """
    # Step 1: Species choice
    if species is None:
        species = _configure_species(species)
    else:
        print(f"Species chosen: {species}")
    species_folder = join(CONFIGURATION_FOLDER, species)

    # Step 2: state and cell types choice
    configuration = parse_configuration_file(
        join(species_folder, f"{species}_cerebellar_cortex.yaml")
    ).__tree__()

    config_cell_types = load_configs_in_folder(join(species_folder, "cell_types"))
    state, cell_types, microzones = _configure_cell_types(
        species_folder, config_cell_types, microzones
    )
    configuration = _update_cell_types(configuration, cell_types, config_cell_types)
    state_folder = join(species_folder, state)
    if microzones:
        micro_params = MicrozonesParams(
            cell_types=["purkinje_cell", "dcn_p", "dcn_i", "io"], labels=["plus", "minus"]
        )
        configuration = _add_microzones(configuration, micro_params)
    else:
        micro_params = MicrozonesParams()

    # Step 3: Simulation choice
    config_simulations = {
        simulator: {
            "cell_models": load_configs_in_folder(join(state_folder, simulator, "cell_models")),
            "connection_models": load_configs_in_folder(
                join(state_folder, simulator, "connection_models")
            ),
            "simulations": load_configs_in_folder(join(state_folder, simulator), recursive=False),
        }
        for simulator in get_folders_in_folder(state_folder)
    }
    simulation_names = _configure_simulations(config_simulations)

    # Step 4: Simulation models choice
    dict_sim, sim_choices = _configure_sim_params(
        config_simulations, simulation_names, micro_params
    )
    deep_update(configuration, dict_sim)

    # Step 5: remove unnecessary cells and connections
    configuration = _clear_unnecessary_params(configuration)

    # Step 6: Add stimulus simulation
    if species == "mouse":
        sim_names = list(configuration["simulations"].keys())
        for sim_name in sim_names:
            simulator, _ = sim_name.split("_", 1)
            if simulator == "nest":
                default_stim = {
                    "mf_stimulus": {
                        "device": "poisson_generator",
                        "rate": 150,
                        "start": 1200,
                        "stop": 1260,
                        "targetting": {
                            "strategy": "sphere",
                            "radius": 90,
                            "origin": [150, 65, 100],
                            "cell_models": ["mossy_fibers"],
                        },
                        "weight": 1,
                        "delay": 0.1,
                    }
                }
                simulation_name = "nest_mf_stimulus"
                if "io" in cell_types:
                    default_stim["mf_stimulus"] = {
                        "device": "poisson_generator",
                        "rate": 40,
                        "start": 1000,
                        "stop": 1260,
                        "targetting": {"strategy": "cell_model", "cell_models": ["mossy_fibers"]},
                        "weight": 1,
                        "delay": 0.1,
                    }
                    default_stim["cf_stimulus"] = {
                        "device": "poisson_generator",
                        "rate": 500,
                        "start": 1250,
                        "stop": 1260,
                        "targetting": {"strategy": "cell_model", "cell_models": ["io"]},
                        "receptor_type": 1,
                        "weight": 55 if state == "vitro" else 100.0,
                        "delay": 0.1,
                    }
                    simulation_name = "nest_mf_cf_stimulus"
                elif "dcn" in cell_types:
                    default_stim["mf_stimulus"]["targetting"]["origin"][2] = 300

                configuration["simulations"][simulation_name] = copy.deepcopy(
                    configuration["simulations"][sim_name]
                )
                deep_update(configuration["simulations"][simulation_name]["devices"], default_stim)
                sim_choices[simulation_name] = sim_choices[sim_name]
            else:  # pragma: no cover
                raise ValueError(
                    f"Only nest configurations are implemented. Provided simulator: {simulator}"
                )
    else:  # pragma: no cover
        raise ValueError(f"Only mouse configuration are implemented. Provided species: {species}")

    # Step 7: Recap choices
    print("\n\nYour choices are:")
    print(f"Species: {species}")
    print(f"State: {state}")
    print(f"Cell types: {cell_types}")
    print(f"With microzones: {microzones}")
    if microzones:
        print(f"\tCell types: {micro_params.cell_types}")
        print(f"\tLabels: {micro_params.labels}")
    print("Simulations:")
    for simulation, choices in sim_choices.items():
        print(f"\t{simulation}:")
        print(f"\t\tCell model: {choices[0]}")
        print(f"\t\tSynapse model: {choices[1]}")
    print("----------------------------\n")
    # Step 8: output folder and extension
    _write_config(configuration, output_folder, extension)



import numpy as np
from bsb import AfterPlacementHook, config, refs, types


@config.node
class LabelMicrozones(AfterPlacementHook):
    """
    Subdivide cell populations into labelled subpopulations of
    same cell counts based on their position along a provided axis.
    The number of labels defines the number of subpopulations
    """

    cell_types: str = config.reflist(refs.cell_type_ref, required=True)
    """Reference to the cell type."""

    axis: int = config.attr(type=types.int(min=0, max=2), default=0)
    """Axis along which to subdivide the population."""

    labels: list[str] = config.list(type=str, default=["type1", "type2"])
    """List of labels to assign to each subpopulation."""

    def postprocess(self):
        for cell_type in self.cell_types:
            # Load the cell type positions
            ps = self.scaffold.get_placement_set(cell_type)
            cell_positions = ps.load_positions()

            # create a filter that split the cells according to
            # the mean of their positions along the chosen axis
            index_pos = np.argsort(cell_positions[:, self.axis])
            split_indexes = np.asarray(
                np.round(np.linspace(0, len(index_pos), len(self.labels) + 1))[1:], dtype=int
            )
            last_i = 0
            for i, label in zip(split_indexes, self.labels):
                subpopulation_1 = np.asarray(index_pos[last_i:i], dtype=int)
                # set the cell label according to the filter
                ps.label(labels=[label], cells=subpopulation_1)
                last_i = i



"""
Module for the configuration node of the Glomerulus to UBC ConnectionStrategy
"""

import numpy as np
from bsb import ConfigurationError, ConnectionStrategy, config

from cerebellar_models.connectome.presyn_dist_strat import PresynDistStrat


@config.node
class ConnectomeGlomerulusUBC(PresynDistStrat, ConnectionStrategy):
    """
    BSB Connection strategy to connect any type of Glomerulus to UBC cells.
    """

    ratios_ubc: dict[str, float] = config.dict(type=float, required=True)
    """Dictionary that links a postsynaptic celltype name to the ratios of the presynaptic UBC 
        population that connects to it."""

    def boot(self):
        parsed_ratios = {
            k.name: (0.0 if k.name not in self.ratios_ubc else self.ratios_ubc[k.name])
            for k in self.presynaptic.cell_types
        }
        sum_ = np.nansum(list(parsed_ratios.values()))
        if np.any(np.array(list(parsed_ratios.values())) < 0):
            raise ConfigurationError("Presynaptic cell type ratios should be greater than 0")
        if sum_ == 0:
            raise ConfigurationError("At least one presynaptic ratio should be greater than 0")
        for k, v in parsed_ratios.items():
            parsed_ratios[k] = v / sum_
        self.ratios_ubc = parsed_ratios.copy()

    def connect(self, pre, post):
        connected_gloms = {
            pre_ps.cell_type.name: np.full(len(pre_ps), False, dtype=bool)
            for pre_ps in pre.placement
        }

        for post_ps in post.placement:
            ubc_pos = post_ps.load_positions()
            ubc_ids = np.random.permutation(len(ubc_pos))
            ubc_pos = ubc_pos[ubc_ids]
            cum_sum = 0
            loc_ratio = 0
            for pre_ps in pre.placement:
                # select the ratio of random ubc ids to connect
                pre_ct = pre_ps.cell_type.name
                loc_ratio += self.ratios_ubc[pre_ct]
                new_ptr = min(len(ubc_ids), int(np.round(len(ubc_pos) * loc_ratio)))
                loc_ubc_ids = ubc_ids[cum_sum:new_ptr]
                cum_sum = new_ptr

                glom_pos = pre_ps.load_positions()
                selected_ubc = ubc_pos[loc_ubc_ids]

                n_conn = len(selected_ubc)  # one glom per ubc
                pre_locs = np.full((n_conn, 3), -1, dtype=int)
                post_locs = np.full((n_conn, 3), -1, dtype=int)
                for i, ubc in enumerate(selected_ubc):
                    filter_connected = ~connected_gloms[pre_ct]
                    if not np.any(filter_connected):
                        # if no glom is left un-connected
                        # we reset the connected list
                        connected_gloms[pre_ct][:] = False
                        filter_connected = ~connected_gloms[pre_ct]
                    avaiable_gloms_ids = np.where(filter_connected)[0]
                    distances = np.linalg.norm(ubc - glom_pos[filter_connected], axis=1)
                    # Take closest glom
                    pre_id = avaiable_gloms_ids[np.argsort(distances)[0]]
                    post_locs[i, 0] = loc_ubc_ids[i]
                    pre_locs[i, 0] = pre_id
                    connected_gloms[pre_ct][pre_id] = True

                self.connect_cells(pre_ps, post_ps, pre_locs, post_locs)



"""
Module for the Report class.
"""

from typing import List, Union
from warnings import warn

import matplotlib.backends.backend_pdf
import numpy as np
from bsb import Scaffold, from_storage

from cerebellar_models.analysis.plots import Plot, ScaffoldPlot


class PlotTypeInfo:
    """
    Class storing information about element plotted, usually cell types or fibers.
    """

    def __init__(self, name: str, color, abbreviation: str = None):
        self.name = name
        """Name of the element."""
        self.abbreviation = abbreviation or name
        """Abbreviation of the element to print in the plot."""
        self.color = color
        """Color of the element."""


LIST_CT_INFO = [
    PlotTypeInfo("mossy_fibers", [0.847, 0, 0.451, 1.0], "mf"),
    PlotTypeInfo("glomerulus", [0.847, 0, 0.451, 1.0], "glom"),
    PlotTypeInfo("granule_cell", [0.7, 0.15, 0.15, 0.5], "GrC"),
    PlotTypeInfo("ascending_axon", [0.7, 0.15, 0.15, 0.5], "aa"),
    PlotTypeInfo("parallel_fiber", [0.7, 0.15, 0.15, 0.5], "pf"),
    PlotTypeInfo("unipolar_brush_cell", [0.196, 0.808, 0.988, 1.0], "UBC"),
    PlotTypeInfo("ubc_glomerulus", [0.196, 0.808, 0.988, 1.0], "ubc_glom"),
    PlotTypeInfo("golgi_cell", [0, 0.45, 0.7, 1.0], "GoC"),
    PlotTypeInfo("purkinje_cell", [0.275, 0.800, 0.275, 1.0], "PC"),
    PlotTypeInfo("basket_cell", [1, 0.647, 0, 1.0], "BC"),
    PlotTypeInfo("stellate_cell", [1, 0.84, 0, 1.0], "SC"),
    PlotTypeInfo("dcn_p", [0.3, 0.3, 0.3, 1.0], "DCN_P"),
    PlotTypeInfo("dcn_i", [0.635, 0, 0.145, 1.0], "DCN_I"),
    PlotTypeInfo("io", [0.46, 0.376, 0.54, 1.0], "IO"),
]
"""Cell and fiber information for plotting cerebellum BSB circuit."""


class Report:
    """
    Class interfacing a list of matplotlib plots to save in an external file.
    """

    def __init__(self, cell_types_info: List[PlotTypeInfo] = None):
        self.plots = {}
        """Dictionary mapping the report's plots' name to their instance"""
        self.cell_types_info = cell_types_info or []
        """List of PlotTypeInfo for each element to plot."""

    @property
    def colors(self):
        """
        Dictionary from the name of the elements to plot to its color.
        """
        return {ct.name: ct.color for ct in self.cell_types_info}

    @property
    def abbreviations(self):
        """
        Dictionary from the name of the elements to plot to its abbreviation.
        """
        return {ct.name: ct.abbreviation for ct in self.cell_types_info}

    def set_plot_colors(self, plot: Plot):
        """
        Set the plot color dictionary according to the report's
        """
        for k, v in self.colors.items():
            plot.set_color(k, v)

    def set_color(self, key: str, color: np.ndarray[float]):
        """
        Set a color for an element to plot.
        Colors must be an array of type RGB or RGBA.
        """
        if len(color) != 3 and len(color) != 4:
            raise ValueError("Color must be an array of size 3 or 4.")
        info_names = [i.name for i in self.cell_types_info]
        new_color = np.array(color, dtype=float)
        if key in info_names:
            self.cell_types_info[info_names.index(key)].color = new_color
        else:
            self.cell_types_info.append(PlotTypeInfo(key, new_color, key))
        for plot in self.plots.values():
            self.set_plot_colors(plot)

    def add_plot(self, name: str, plot: Plot):
        """
        Add a plot to the list of the report's plots.
        """
        if name in list(self.plots.keys()):
            warn("A plot named '{}' already exists in the report. Skipping it".format(name))
            return
        self.set_plot_colors(plot)
        self.plots[name] = plot

    def preprocessing(self):
        """
        Function to apply modifications to the report's plots
        after the plotting is done.
        """
        pass

    def print_report(self, output_name: str, dpi: int = 200, pad: int = 0, **kwargs):
        """
        Print the report and export it in a pdf file.
        """
        pdf = matplotlib.backends.backend_pdf.PdfPages(output_name)
        self.preprocessing()
        for name, plot in self.plots.items():
            if not plot.is_plotted:
                plot.plot()
            plot.figure.tight_layout(pad=pad)
            plot.figure.savefig(pdf, format="pdf", dpi=dpi, **kwargs)
        pdf.close()

    def save_plot(self, plot_name: str, output_name: str, dpi: int = 200, pad: int = 0, **kwargs):
        """
        Save one of the report's plots as a separate figure.
        """
        if plot_name in self.plots:
            self.plots[plot_name].save_figure(output_name, dpi, pad, **kwargs)

    def show_plot(self, plot_name):  # pragma:nocover
        """
        Show one of the report's plots.
        """
        if plot_name in self.plots:
            self.plots[plot_name].show()

    def show(self):  # pragma:nocover
        """
        Show all report's plots one by one.
        """
        for plot in self.plots.values():
            plot.show()


class BSBReport(Report):
    """
    Class interfacing a list of matplotlib plots for analysis of BSB reconstructions.
    """

    def __init__(self, scaffold: Union[str, Scaffold], cell_types_info: List[PlotTypeInfo] = None):
        super().__init__(cell_types_info or LIST_CT_INFO)
        if isinstance(scaffold, Scaffold):
            self.scaffold = scaffold
        else:
            self.scaffold = from_storage(scaffold)

    @property
    def cell_names(self):
        """
        List of the name of the elements to plot.
        """
        return list(self.scaffold.cell_types.keys())

    @property
    def labelled_cell_names(self):
        ct_names = []
        for ps in self.scaffold.get_placement_sets():
            ct_name = ps.cell_type.name
            for labels in ps.get_unique_labels():
                ct_names.append(ScaffoldPlot.get_labelled_ct_name(ct_name, labels))
        return ct_names

    @property
    def labelled_abbreviations(self):
        dict_abv = self.abbreviations.copy()
        for ct_name in self.labelled_cell_names:
            if ct_name not in self.cell_names:
                ct, label = ct_name.rsplit("_", 1)
                dict_abv[ct_name] = self.abbreviations[ct] + "_" + label
        return dict_abv

    def add_plot(self, name: str, plot: Plot):
        super().add_plot(name, plot)
        if isinstance(plot, ScaffoldPlot):
            plot.set_scaffold(self.scaffold)



"""
Module for the configuration node of the Glomerulus to Golgi ConnectionStrategy
"""

import itertools

import numpy as np
from bsb import ConnectionStrategy, ConnectivityError, config
from scipy.stats.distributions import truncexpon

from cerebellar_models.connectome.presyn_dist_strat import PresynDistStrat


@config.node
class ConnectomeGlomerulusGolgi(PresynDistStrat, ConnectionStrategy):
    """
    BSB Connection strategy to connect Glomerulus to Golgi cells.
    """

    def connect(self, pre, post):
        for pre_ps in pre.placement:
            for post_ps in post.placement:
                self._connect_type(pre_ps, post_ps)

    def _connect_type(self, pre_ps, post_ps):
        glomeruli_pos = pre_ps.load_positions()
        golgi_pos = post_ps.load_positions()

        # If synaptic contacts need to be made we use this exponential distribution
        # to pick the closer by subcell_labels.
        exp_dist = truncexpon(b=5, scale=0.03)
        golgi_morphos = post_ps.load_morphologies().iter_morphologies(cache=True, hard_cache=True)

        n_conn = len(glomeruli_pos) * len(golgi_pos)
        pre_locs = np.full((n_conn, 3), -1, dtype=int)
        post_locs = np.full((n_conn, 3), -1, dtype=int)
        ptr = 0
        for i, golgi, morpho in zip(itertools.count(), golgi_pos, golgi_morphos):
            to_connect_idx = np.nonzero(
                np.linalg.norm(golgi - glomeruli_pos, axis=1) < self.radius
            )[0]
            connected_gloms = len(to_connect_idx)
            pre_locs[ptr : ptr + connected_gloms, 0] = to_connect_idx
            post_locs[ptr : ptr + connected_gloms, 0] = i

            # Find terminal points on branches
            basal_dendrides_branches = morpho.get_branches()
            terminal_branches_ids = np.nonzero([b.is_terminal for b in basal_dendrides_branches])[0]
            basal_dendrides_branches = np.take(
                basal_dendrides_branches, terminal_branches_ids, axis=0
            )
            if basal_dendrides_branches.size == 0:
                raise ConnectivityError(
                    "The golgi morphology provided has no terminal branches.\n"
                    "Check the morphology_labels."
                )

            # Find the point-on-branch ids of the tips
            tips_coordinates = np.array([b.points[-1] for b in basal_dendrides_branches])

            # Connect each close by glom to a tip of the golgi
            for id_g, glom_p in enumerate(to_connect_idx):
                sorted_pts_ids = np.argsort(
                    np.linalg.norm(tips_coordinates + golgi - glom_p, axis=1)
                )
                # Pick the golgi tip according to an exponential distribution mapped
                # through the distance to each glom: high chance to pick close by tips.
                pt_idx = sorted_pts_ids[tips_coordinates.size * exp_dist.rvs(size=1).astype(int)[0]]

                post_locs[ptr + id_g, 1] = morpho.branches.index(basal_dendrides_branches[pt_idx])
                post_locs[ptr + id_g, 2] = len(basal_dendrides_branches[pt_idx]) - 1
            ptr += connected_gloms

        self.connect_cells(pre_ps, post_ps, pre_locs[:ptr], post_locs[:ptr])



import shutil
from os import makedirs
from os.path import abspath, dirname, exists, isdir, join

import appdirs
from bsb.services import MPI
from pynestml.frontend.pynestml_frontend import generate_target

_cereb_dirs = appdirs.AppDirs("cerebellar_models")
_cache_path = _cereb_dirs.user_cache_dir


def _build_nest_models(
    model_dir=dirname(__file__),
    build_dir=join(_cache_path, "nest_build"),
    module_name="cerebmodule",
    redo=False,
):
    """
    Build all the nestml models within the provided model directory and deploy them.

    :param str model_dir: Directory containing the nestml files to compile
    :param str build_dir: Directory where the nest models will be compiled
    :param str module_name: Name of the nest module produced as outcome.
    :param bool redo: Flag to force rebuild the nest models.
    """

    model_dir = abspath(model_dir)
    if MPI.get_size() == 1 or MPI.get_rank() == 0:
        if not redo:
            import nest

            nest.ResetKernel()
            try:
                nest.Install(module_name)
                # unload the module
                nest.ResetKernel()
                return
            except nest.NESTErrors.DynamicModuleManagementError as e:
                if "loaded already" in e.message:
                    return
        if not (exists(model_dir) and isdir(model_dir)):
            raise OSError("Model directory does not exist: {}".format(model_dir))
        if exists(build_dir) and isdir(build_dir):
            shutil.rmtree(build_dir)
        makedirs(build_dir)

        generate_target(
            input_path=model_dir,
            target_platform="NEST",
            target_path=build_dir,
            module_name=module_name,
        )


_build_nest_models()



"""
Module for the configuration node of every presynaptic cell to Glomerulus ConnectionStrategy
"""

import abc
import itertools

import numpy as np
from bsb import Chunk, ConnectionStrategy, InvertedRoI, config

from cerebellar_models.connectome.presyn_dist_strat import PresynDistStrat


def norm_exp_dist(size: int = 1, b: float = 2.0):
    """
    Normalized exponential random generator for distance based selection

    :param int size: number of random to sample
    :param float b: strength of the exponential decay.
    :return: random numbers sampled
    :rtype: numpy.ndarray
    """
    return (np.exp(-b * np.random.rand(size)) - np.exp(-b)) / (1 - np.exp(-b))


class ConnectomeGlomerulus(InvertedRoI, ConnectionStrategy):
    """
    BSB Connection strategy to connect a presynaptic cell to Glomeruli.
    """

    def connect(self, pre, post):
        # We use a truncated exponential distribution to favour the presynaptic fibers closer the
        # postsynaptic glomerulus.
        for pre_ps in pre.placement:
            for post_ps in post.placement:
                presyn_pos = pre_ps.load_positions()
                glomeruli_pos = post_ps.load_positions()
                n_glom = len(glomeruli_pos)

                # We work assuming that there is at least 1 presynaptic fiber in the ROI.
                # Otherwise, there is something wrong in the placement phase.
                pre_locs = np.full((n_glom, 3), -1, dtype=int)
                post_locs = np.full((n_glom, 3), -1, dtype=int)

                # We connect each glomerulus to a presynaptic cell.
                to_keep = 0
                for j, glomerulus in enumerate(glomeruli_pos):
                    pre_ids = self.pre_selection(presyn_pos, glomerulus)
                    if len(pre_ids) > 0:
                        roll = int(np.floor(len(pre_ids) * norm_exp_dist()[0]))
                        pre_locs[to_keep, 0] = pre_ids[roll]
                        post_locs[to_keep, 0] = j
                        to_keep += 1

                self.connect_cells(pre_ps, post_ps, pre_locs[:to_keep], post_locs[:to_keep])

    @abc.abstractmethod
    def pre_selection(
        self,
        presyn_pos,
        glom_pos,
    ):  # pragma: no cover
        """
        Order presynaptic cell ids based on their respective distance to glomerulus

        :param numpy.ndarray presyn_pos: list of presynaptic cell positions
        :param numpy.ndarray glom_pos: single glomerulus position
        :return: presynaptic cell ids sorted by distance to glomerulus
        :rtype: numpy.ndarray
        """
        pass


@config.node
class ConnectomeMossyGlomerulus(ConnectomeGlomerulus):
    """
    BSB Connection strategy to connect Mossy fibers to Glomeruli.
    """

    x_length: float = config.attr(type=float, required=True)
    """Length of the box along the x axis surrounding the glomerulus cell soma in which the 
        presynaptic cell can be connected."""
    y_length: float = config.attr(type=float, required=True)
    """Length of the box along the y axis surrounding the glomerulus cell soma in which the 
        presynaptic cell can be connected."""

    def pre_selection(
        self,
        presyn_pos,
        glom_pos,
    ):
        diff = np.absolute(glom_pos - presyn_pos)
        ids_to_keep = np.where((diff[:, 0] <= self.x_length) * (diff[:, 1] <= self.y_length))[0]
        dist = np.linalg.norm(diff[ids_to_keep], axis=1)
        return ids_to_keep[np.argsort(dist)]

    def get_region_of_interest(self, chunk):
        # Chunk here is a postsynaptic chunk because of InvertedRoI
        # We look for chunks containing mossy fibers that are within a rectangle of size
        # x_length * y_length centered on the postsynaptic chunk containing the glomerulus.
        chunks = set(
            itertools.chain.from_iterable(
                ct.get_placement_set().get_all_chunks() for ct in self.presynaptic.cell_types
            )
        )
        selected_chunks = []
        for c in chunks:
            x_dist = np.fabs(chunk[0] - c[0]) * chunk.dimensions[0]
            y_dist = np.fabs(chunk[1] - c[1]) * chunk.dimensions[1]

            if (x_dist < self.x_length / 2) and (y_dist < self.y_length / 2):
                selected_chunks.append(Chunk([c[0], c[1], c[2]], chunk.dimensions))
        return selected_chunks


@config.node
class ConnectomeUBCGlomerulus(ConnectomeGlomerulus, PresynDistStrat):
    """
    BSB Connection strategy to connect UBC to Glomeruli.
    """

    def pre_selection(
        self,
        presyn_pos,
        glom_pos,
    ):
        dist = np.linalg.norm(presyn_pos - glom_pos, axis=1)
        ids_to_keep = np.where(dist <= self.radius)[0]
        return ids_to_keep[np.argsort(dist[ids_to_keep])]



"""
Module for the plots and reports related to the simulation analysis of BSB scaffold.
"""

from os import listdir
from os.path import isfile, join
from typing import List, Tuple, Union

import numpy as np
from bsb import Scaffold
from elephant.kernels import GaussianKernel, Kernel
from elephant.statistics import instantaneous_rate, isi
from matplotlib import gridspec as gs
from matplotlib import pyplot as plt
from neo import SpikeTrain
from neo import io as nio
from quantities import ms

from cerebellar_models.analysis.plots import Legend, Plot, ScaffoldPlot
from cerebellar_models.analysis.report import BSBReport, PlotTypeInfo
from cerebellar_models.analysis.structure_analysis import TablePlot


def _check_simulation(scaffold: Scaffold, simulation_name: str):
    """
    Check if a simulation is in a Scaffold and raise an error if not.
    """
    if simulation_name not in scaffold.simulations:
        raise ValueError(f"Simulation name {simulation_name} not in the scaffold simulations")


class SpikePlot(ScaffoldPlot):
    """
    Abstract class for plotting the spiking simulation results of a BSB scaffold.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons: List,
        populations: List,
        dict_colors: dict = None,
        **kwargs,
    ):
        _check_simulation(scaffold, simulation_name)
        super().__init__(fig_size, scaffold, dict_colors=dict_colors, **kwargs)
        self.simulation_name = simulation_name
        """Name of the simulation as defined in the scaffold configuration."""
        self._time_from = time_from or 0
        """Start time of the analysis"""
        self.time_to = time_to or self.scaffold.simulations[self.simulation_name].duration
        """End time of the analysis. By default, this corresponds to the simulation duration."""
        self.dt = self.scaffold.simulations[simulation_name].resolution
        """Time step of the simulation in ms"""
        self.all_spikes = all_spikes
        """List of SpikeTrain for each cell type"""
        if len(nb_neurons) != len(populations):
            raise ValueError("populations and nb_neurons must have the same length")
        self.nb_neurons = nb_neurons
        """Number of neuron for each neuron type"""
        self.populations = populations
        """List of neuron type names"""

    def _check_times(self, start, stop):
        if stop < 0 or start < 0:
            raise ValueError("time_from and time_to must be non-negative")
        max_time = self.scaffold.simulations[self.simulation_name].duration
        if stop > max_time:
            raise ValueError("time_to must be less than the simulation's duration")
        if start > stop:
            raise ValueError("time_from must be less than time_to")

    @property
    def time_to(self):
        return self._time_to

    @time_to.setter
    def time_to(self, value):
        self._check_times(self.time_from, value)
        self._time_to = value

    @property
    def time_from(self):
        return self._time_from

    @time_from.setter
    def time_from(self, value):
        self._check_times(value, self.time_to)
        self._time_from = value

    def _set_simulation_params(
        self,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons,
        populations,
    ):
        is_different = (
            self.simulation_name != simulation_name
            or self.time_from != time_from
            or self.time_to != time_to
            or np.any(self.all_spikes != all_spikes)
            or np.any(self.nb_neurons != nb_neurons)
            or self.populations != populations
        )
        if is_different:
            _check_simulation(self.scaffold, simulation_name)
            self.simulation_name = simulation_name
            self.time_from = time_from
            self.time_to = time_to
            self.all_spikes = all_spikes
            self.nb_neurons = nb_neurons
            self.populations = populations
            self.is_updated = False
            if self.is_plotted:
                self.clear()
        return is_different

    def get_filt_spikes(self):
        """
        Filter the spike events for the time of the analysis.

        :return: Sliced List of SpikeTrain.
        :rtype: List[neo.core.SpikeTrain]
        """
        return [st.time_slice(self.time_from * ms, self.time_to * ms) for st in self.all_spikes]


class SpikeSimulationReport(BSBReport):
    """
    Abstract class for reports of simulation results of BSB scaffold.
    """

    def __init__(
        self,
        scaffold: Union[str, Scaffold],
        simulation_name: str,
        folder_nio: str,
        time_from: float = 0,
        time_to: float = None,
        ignored_ct=None,
        cell_types_info: List[PlotTypeInfo] = None,
    ):
        super().__init__(scaffold, cell_types_info)
        _check_simulation(self.scaffold, simulation_name)
        self.simulation_name = simulation_name
        """Name of the simulation as defined in the scaffold configuration."""
        self.folder_nio = folder_nio
        """Folder containing the simulation results stored as nio files."""
        self._time_from = time_from
        """Start time of the analysis"""
        self.time_to = time_to or self.scaffold.simulations[simulation_name].duration
        """End time of the analysis. By default, this corresponds to the simulation duration."""
        self.dt = self.scaffold.simulations[simulation_name].resolution
        """Time step of the simulation in ms"""
        self.ignored_ct = ignored_ct or ["glomerulus", "ubc_glomerulus"]
        """List of ignored cell type names"""
        self.all_spikes: List[SpikeTrain] = None
        """List of SpikeTrain for each cell type"""
        self.nb_neurons = None
        """Number of neuron for each neuron type"""
        self.populations = None
        """List of neuron type names"""

        self.all_spikes, self.nb_neurons, self.populations = self.load_spikes()

    def _check_times(self, start, stop):
        if stop < 0 or start < 0:
            raise ValueError("time_from and time_to must be non-negative")
        max_time = self.scaffold.simulations[self.simulation_name].duration
        if stop > max_time:
            raise ValueError("time_to must be less than the simulation's duration")
        if start > stop:
            raise ValueError("time_from must be less than time_to")

    @property
    def time_to(self):
        return self._time_to

    @time_to.setter
    def time_to(self, value):
        self._check_times(self.time_from, value)
        self._time_to = value
        for plot in self.plots.values():
            if isinstance(plot, SpikePlot):
                plot.time_to = value

    @property
    def time_from(self):
        return self._time_from

    @time_from.setter
    def time_from(self, value):
        self._check_times(value, self.time_to)
        self._time_from = value
        for plot in self.plots.values():
            if isinstance(plot, SpikePlot):
                plot.time_from = value

    def _extract_ct_device_name(self, device_name: str):
        """Extract the cell type name from its device name."""
        if "_record" in device_name:
            targetting = (
                self.scaffold.simulations[self.simulation_name].devices[device_name].targetting
            )
            ct = targetting.cell_models_references[0]
            labels = targetting["labels"] if "labels" in targetting else set()
            return ct, labels
        else:
            return device_name, set()

    def _extract_spikes_dict(self):
        """
        Extract the spike events from nio files stored in a folder and group them by neuron type.

        :return: - List of spike events grouped by neuron type.
                 - Dictionary storing for each neuron type its index and its unique list of neuron ids.
                   The index is stored under the "id" key and the neuron ids are stored under the "senders" key.
        :rtype: Tuple[List[List[float]], Dict[str, numpy.ndarray[int]]
        """
        spikes_res = []
        cell_dict = {}
        current_id = 0

        for f in listdir(self.folder_nio):
            file_ = join(self.folder_nio, f)
            if isfile(file_) and (".nio" in file_):
                block = nio.NixIO(file_, mode="ro").read_all_blocks()[0]  # assume only one block
                spiketrains = block.segments[0].spiketrains  # assume only one segment

                for st in spiketrains:
                    st.segment = None  # remove spiketrain segment to allow merging
                    cell_type, labels = self._extract_ct_device_name(st.annotations["device"])
                    if cell_type in self.cell_names and cell_type not in self.ignored_ct:
                        cell_type_label = ScaffoldPlot.get_labelled_ct_name(cell_type, labels)
                        if cell_type_label not in cell_dict:
                            cell_dict[cell_type_label] = current_id
                            current_id += 1
                            spikes_res.append([])
                        spikes_res[cell_dict[cell_type_label]].append(st)
        return spikes_res, cell_dict

    def load_spikes(self):
        """
        Load the spike trains from nio files.

        :return: - Boolean numpy array of shape (N*M) storing spike events for each time step.
                   N corresponds to the number of time steps, M to the number of neuron. Neurons are sorted by type.
                 - List of number of unique neuron per type.
                 - List of cell type names.
        :rtype: Tuple[List[neo.core.SpikeTrain], numpy.ndarray[int], List[str]]
        """
        spikes_res, cell_dict = self._extract_spikes_dict()
        all_spikes = []
        nb_neurons = np.zeros(len(cell_dict), dtype=int)
        for i, cell_type in enumerate(cell_dict):
            sts = spikes_res[cell_dict[cell_type]]
            merged = sts[0]
            for st in sts[1:]:
                merged = merged.merge(st)
            all_spikes.append(merged)
            nb_neurons[i] = all_spikes[i].annotations["pop_size"]
        return all_spikes, nb_neurons, list(cell_dict.keys())

    def add_plot(self, name: str, plot: Plot):
        super().add_plot(name, plot)
        if isinstance(plot, SpikePlot):
            plot._set_simulation_params(
                self.simulation_name,
                self.time_from,
                self.time_to,
                self.all_spikes,
                self.nb_neurons,
                self.populations,
            )

    def get_filt_spikes(self):
        """
        Filter the spike events for the time of the analysis.

        :return: Sliced List of SpikeTrain.
        :rtype: List[neo.core.SpikeTrain]
        """
        return [st.time_slice(self.time_from * ms, self.time_to * ms) for st in self.all_spikes]


class RasterPSTHPlot(SpikePlot):
    """
    Combined raster plot and PSTH plot of the spiking activity results for each neuron type.
    The subplots are split in two columns.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons: List,
        populations: List,
        nb_bins: int = 30,
        dict_colors: dict = None,
        **kwargs,
    ):
        # population needs to be set before the super.__init__ because it is used in init_plot
        self.populations = populations
        super().__init__(
            fig_size,
            scaffold,
            simulation_name,
            time_from,
            time_to,
            all_spikes,
            nb_neurons,
            populations,
            dict_colors,
            **kwargs,
        )
        if nb_bins <= 0:
            raise ValueError("nb_bins must be greater than 0.")
        self.nb_bins = nb_bins
        """Number of bins for the PSTH subplot."""

    def init_plot(self, **kwargs):
        if self.is_initialized:
            plt.close(self.figure)
        self.is_initialized = True
        self.is_plotted = False
        self.nb_cols = 2
        num_filter = len(self.populations)
        self.nb_rows = int(np.ceil(num_filter / 2.0))  # nb rows
        self.figure = plt.figure(figsize=self.fig_size, **kwargs)
        if self.nb_rows > 0:
            global_gsp = gs.GridSpec(self.nb_rows, 2)
        self.axes = [[] for _ in range(self.nb_rows)]
        for i in range(num_filter):
            local_gsp = gs.GridSpecFromSubplotSpec(2, 1, subplot_spec=global_gsp[i], hspace=0)
            ax1 = plt.Subplot(self.figure, local_gsp[0])
            self.figure.add_subplot(ax1)

            ax2 = plt.Subplot(self.figure, local_gsp[1])
            self.figure.add_subplot(ax2)
            self.axes[i // 2].append([ax1, ax2])

    def clear(self):
        for ax in self.get_axes():
            ax[0].clear()
            ax[1].clear()
        self.is_plotted = False

    def plot(self, relative_time=False, params_raster: dict = None, params_psth: dict = None):
        """
        Plot or replot the figure
        Calls the update function if needed.

        :param bool relative_time: If True, the x-axis values will be relative to the time interval.
        :param params_raster: Dictionary of parameters for the raster plot (see matplotlib scatter).
        :param params_psth: Dictionary of parameters for the PSTH plot (see matplotlib hist).
        """
        super().plot()

        # extract dict params
        loc_params_raster = {"marker": "o", "alpha": 1, "rasterized": True}
        if params_raster is not None:
            loc_params_raster.update(params_raster)
        params_psth = params_psth if params_psth is not None else {}

        num_filter = len(self.nb_neurons)
        counts = np.zeros(num_filter + 1)
        counts[1:] = np.cumsum(self.nb_neurons)

        bin_times = np.linspace(self.time_from, self.time_to, self.nb_bins)
        loc_spikes = self.get_filt_spikes()
        for i, ct in enumerate(self.populations):
            times = loc_spikes[i].magnitude
            _, newIds = np.unique(loc_spikes[i].array_annotations["senders"], return_inverse=True)
            cell_params = loc_params_raster.copy()
            if "s" not in cell_params and self.nb_neurons[i] > 0:
                cell_params["s"] = 50.0 / self.nb_neurons[i]
            color = (
                self.labelled_dict_colors[ct][:3]
                if ct in self.labelled_dict_colors
                else [0.6, 0.6, 0.6]
            )
            ax = self.get_ax(i)[0]
            if self.nb_neurons[i] > 0:
                ax.scatter(
                    times,
                    newIds,
                    color=color,
                    **cell_params,
                )
            ax.invert_yaxis()
            ax.set_xlim(
                [0, self.time_to - self.time_from]
                if relative_time
                else [self.time_from, self.time_to]
            )
            ax.get_xaxis().set_visible(False)
            ax.set_ylabel("Neuron id")
            ax.set_title(f"{ct}")

            ax = self.get_ax(i)[1]
            if self.nb_neurons[i] > 0:
                ax.hist(times, bin_times, color=color, **params_psth)
            ax.set_xlabel("Time in ms")
            ax.set_xlim(
                [0, self.time_to - self.time_from]
                if relative_time
                else [self.time_from, self.time_to]
            )
            ax.set_ylabel("Spike counts")


class Spike2Columns(SpikePlot):
    """
    Utility class to plot simulation results for each neuron type in a 2 columns fashion.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons: List,
        populations: List,
        dict_colors: dict = None,
        **kwargs,
    ):
        # population needs to be set before the super.__init__ because it is used in init_plot
        self.populations = populations
        super().__init__(
            fig_size,
            scaffold,
            simulation_name,
            time_from,
            time_to,
            all_spikes,
            nb_neurons,
            populations,
            dict_colors,
            **kwargs,
        )

    def init_plot(self, **kwargs):
        if self.is_initialized:
            plt.close(self.figure)
        self.is_initialized = True
        self.is_plotted = False
        self.nb_cols = 2
        num_filter = len(self.populations)
        self.nb_rows = int(np.ceil(num_filter / 2.0))  # nb rows
        self.figure = plt.figure(figsize=self.fig_size, **kwargs)
        self.axes = [[] for _ in range(self.nb_rows)]
        for i in range(num_filter):
            self.axes[i // 2].append(
                plt.subplot2grid((self.nb_rows, 2), (i // 2, i % 2), rowspan=1, fig=self.figure)
            )


class FiringRatesPlot(Spike2Columns):
    """
    Instantaneous firing rate plot for each cell type based on a time kernel.
    Each population firing rate signal is plotted surrounding by its standard deviation
    A firing rate signal is computed as the mean of the convolution of spike times
    for each neuron with the time kernel.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons: List,
        populations: List,
        kernel=None,
        dict_colors: dict = None,
        **kwargs,
    ):
        super().__init__(
            fig_size,
            scaffold,
            simulation_name,
            time_from,
            time_to,
            all_spikes,
            nb_neurons,
            populations,
            dict_colors,
            **kwargs,
        )
        if kernel is not None and not isinstance(kernel, Kernel):
            raise TypeError("Kernel must be an instance of elephant Kernel or None")
        self.kernel = kernel or "auto"
        """Width of the kernel expressed as number of ms"""

    def update(self):
        super().update()
        num_filter = len(self.nb_neurons)
        counts = np.zeros(num_filter + 1)
        counts[1:] = np.cumsum(self.nb_neurons)

        loc_spikes = self.get_filt_spikes()
        duration = int((self.time_to - self.time_from) / self.dt)
        self.firing_rates = np.zeros((duration, num_filter))
        for i in range(num_filter):
            if loc_spikes[i].size <= 0:
                continue  # pragma: nocover
            self.firing_rates[:, i] = (
                instantaneous_rate(
                    loc_spikes[i],
                    sampling_period=self.dt * ms,
                    kernel=self.kernel,
                    border_correction=True,
                ).magnitude[:, 0]
                / self.nb_neurons[i]
            )

    def plot(self, relative_time=False, **kwargs):
        """
        Plot or replot the figure
        Calls the update function if needed.

        :param bool relative_time: If True, the x-axis values will be relative to the time interval.
        """
        super().plot()
        time_interval = np.arange(
            self.time_from,
            self.time_to,
            self.dt,
        )
        for i, ct in enumerate(self.populations):
            ax = self.get_ax(i)
            ax.plot(
                time_interval,
                self.firing_rates[:, i],
                color=self.labelled_dict_colors[ct][:3],
                **kwargs,
            )
            ax.set_xlabel("Time in ms")
            ax.set_ylabel("Rate in Hz")
            kernel_text = (
                f" (kernel width = {self.kernel.sigma})" if isinstance(self.kernel, Kernel) else ""
            )
            ax.set_title(f"Mean estimated firing rate for {ct}{kernel_text}")
            ax.set_xlim(
                [0, time_interval[-1] - time_interval[0]]
                if relative_time
                else [time_interval[0], time_interval[-1]]
            )
            ax.text(
                0.01,
                0.95,
                r"FR: {:.2} $\pm$ {:.2}".format(
                    np.mean(self.firing_rates[:, i]), np.std(self.firing_rates[:, i])
                ),
                ha="left",
                va="top",
                transform=ax.transAxes,
            )


def extract_isis(spikes, dt):
    """
    Extract inter-spike intervals from a list of spike trains.
    One mean inter-spike interval value is computed for each neuron.

    :param neo.core.SpikeTrain spikes: population SpikeTrain
    :param float dt: time step

    :return: list of inter-spike intervals
    :rtype: List[float]
    """

    isi_ = []
    senders = spikes.array_annotations["senders"]
    u_senders, inv = np.unique(senders, return_inverse=True)
    mat = np.zeros((int((spikes.t_stop - spikes.t_start) / dt), len(u_senders)), dtype=bool)
    mat[np.asarray(np.rint((spikes.times - spikes.t_start) / dt), dtype=int) - 1, inv] = True
    for sender in range(len(u_senders)):
        isis = isi(np.where(mat[:, sender])[0] * dt * ms)
        if len(isis) > 0:
            isi_.append(np.mean(isis))
    return isi_


class ISIPlot(Spike2Columns):
    """
    Inter-spike interval histogram plot for each cell type.
    For each neuron type, one mean inter-spike interval value is computed for each of its neuron.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons: List,
        populations: List,
        nb_bins: int = 50,
        dict_colors: dict = None,
        **kwargs,
    ):
        super().__init__(
            fig_size,
            scaffold,
            simulation_name,
            time_from,
            time_to,
            all_spikes,
            nb_neurons,
            populations,
            dict_colors,
            **kwargs,
        )
        if nb_bins <= 0:
            raise ValueError("nb_bins must be greater than 0.")
        self.nb_bins = nb_bins
        """Number of bins of the histogram."""

    def plot(self, **kwargs):
        super().plot()
        num_filter = len(self.nb_neurons)
        counts = np.zeros(num_filter + 1)
        counts[1:] = np.cumsum(self.nb_neurons)
        isis_dist = [extract_isis(self.all_spikes[i], self.dt) for i in range(num_filter)]
        for i, ct in enumerate(self.populations):
            ax2 = self.get_ax(i)
            if len(isis_dist[i]) > 0:
                ax2.hist(
                    isis_dist[i], self.nb_bins, color=self.labelled_dict_colors[ct][:3], **kwargs
                )
            ax2.set_xlabel("ISIs bins in ms")
            ax2.set_yscale("log")
            ax2.set_title(f"Distribution of {ct} ISIs")


class FrequencyPlot(FiringRatesPlot):
    """
    Plot of the frequency distribution analysis of the instantaneous firing rate signal.
    """

    def update(self):
        super().update()
        self.frequencies = np.zeros((self.firing_rates.shape[1], self.firing_rates.shape[0] // 2))
        self.freq_powers = np.zeros((self.firing_rates.shape[1], self.firing_rates.shape[0] // 2))
        for i, fr in enumerate(self.firing_rates.T):
            glob_fr = fr[:-1]
            t = np.abs(np.fft.fft(glob_fr))
            x = np.fft.fftfreq(t.shape[0], self.dt / 1e3)  # convert ms to s
            idx = np.argsort(x)
            self.freq_powers[i] = t[idx][t.shape[0] // 2 :] * 2
            self.frequencies[i] = x[idx][x.shape[0] // 2 :]

    def plot(self, max_freq=30.0, plot_bands=True, **kwargs):
        """
        Plot or replot the figure
        Calls the update function if needed.

        :param float max_freq: maximum frequency (in Hz).
        :param bool plot_bands: if True, plot the frequency bands.
        """
        super(FiringRatesPlot, self).plot()
        dict_plot = {}
        dict_plot.update(kwargs)
        for i, (fr, pw, ct) in enumerate(zip(self.frequencies, self.freq_powers, self.populations)):
            ax = self.get_ax(i)
            ax.plot(fr[1:], pw[1:], color=self.labelled_dict_colors[ct], label=ct, **dict_plot)
            ax.set_xlim([0.0, max_freq])
            ax.set_xlabel("Frequency [Hz]")
            ax.set_ylabel("Power [dB]")
            ax.set_title(f"Frequency spectrum for {ct}")
            ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
            if plot_bands:
                ax.axvline(4.0, ls="--", color="black")
                ax.axvline(8.0, ls="--", color="black")
                ax.axvline(12.0, ls="--", color="black")
                ax.axvline(30.0, ls="--", color="black")


class SimResultsTable(TablePlot, SpikePlot):
    """
    Table of the firing rates and inter-spike intervals for each cell type.
    The firing rate value of a cell type corresponds to the mean number of spike over the time interval,
    while its inter-spike interval corresponds to the mean of all mean inter-spike interval values
    computed for each of its neuron.
    """

    def __init__(
        self,
        fig_size: Tuple[float, float],
        scaffold: Scaffold,
        simulation_name: str,
        time_from: float,
        time_to: float,
        all_spikes,
        nb_neurons: List,
        populations: List,
        dict_colors: dict = None,
        dict_abv=None,
        **kwargs,
    ):
        super().__init__(
            fig_size,
            scaffold,
            simulation_name,
            time_from,
            time_to,
            all_spikes,
            nb_neurons,
            populations,
            dict_colors,
            **kwargs,
        )
        self.columns = ["Firing rate [Hz]", "Inter Spike Intervals [ms]"]
        self.dict_abv = dict_abv or {}
        """Dictionary of abbreviations for cell types"""

    def plot(self, **kwargs):
        super().plot()
        self.plot_table(**kwargs)

    def update(self):
        super().update()
        self.reset_table()
        num_filter = len(self.nb_neurons)
        counts = np.zeros(num_filter + 1)
        counts[1:] = np.cumsum(self.nb_neurons)
        loc_spikes = self.get_filt_spikes()
        for i in range(num_filter):
            spikes = loc_spikes[i]

            all_fr = np.unique(spikes.array_annotations["senders"], return_counts=True)[1] / (
                (self.time_to - self.time_from) / 1000.0
            )
            isi = extract_isis(spikes, self.dt)

            self._values.append([all_fr, isi])
            self.table_values.append(
                [
                    (
                        "{:.2} pm {:.2}".format(np.mean(all_fr), np.std(all_fr))
                        if len(all_fr) > 0
                        else "/"
                    ),
                    "{:.2} pm {:.2}".format(np.mean(isi), np.std(isi)) if len(isi) > 0 else "/",
                ]
            )
        self.rows = [(self.dict_abv[ct] if ct in self.dict_abv else ct) for ct in self.populations]

    def get_firing_rates(self):
        """
        Return a dictionary which gives for each cell type the firing rate
        of each neuron spiking.
        The plot needs to be updated.

        :rtype: Dict[str, int]
        """
        return {ct: line[0] for ct, line in zip(self.rows, self._values)}

    def get_isis_values(self):
        return {ct: line[1] for ct, line in zip(self.rows, self._values)}


class BasicSimulationReport(SpikeSimulationReport):
    """
    Simulation report of the spike activity containing:

    - a plot with the raster and PSTH for each cell type,
    - a table plot storing the mean firing rate and ISI value for each cell type,
    - an instantaneous firing rate plot for each cell type,
    - an inter-spike interval histogram plot for each cell type,
    - a frequency spectrum plot for each cell type,
    - a legend plot
    """

    def __init__(
        self,
        scaffold: Union[str, Scaffold],
        simulation_name: str,
        folder_nio: str,
        time_from: float = 0,
        time_to: float = None,
        ignored_ct: list[str] = [],
        cell_types_info: List[PlotTypeInfo] = None,
    ):
        super().__init__(
            scaffold, simulation_name, folder_nio, time_from, time_to, ignored_ct, cell_types_info
        )
        num_labelled_ct = len(self.populations)
        raster = RasterPSTHPlot(
            (15, 3 * np.ceil(num_labelled_ct / 2)),
            scaffold=self.scaffold,
            simulation_name=self.simulation_name,
            time_from=self.time_from,
            time_to=self.time_to,
            all_spikes=self.all_spikes,
            nb_neurons=self.nb_neurons,
            populations=self.populations,
        )
        table = SimResultsTable(
            (5, 0.22 * (num_labelled_ct + 1)),
            scaffold=self.scaffold,
            simulation_name=self.simulation_name,
            time_from=self.time_from,
            time_to=self.time_to,
            all_spikes=self.all_spikes,
            nb_neurons=self.nb_neurons,
            populations=self.populations,
            dict_abv=self.abbreviations,
        )
        firing_rates = FiringRatesPlot(
            (15, 2 * np.ceil(num_labelled_ct / 2)),
            scaffold=self.scaffold,
            simulation_name=self.simulation_name,
            time_from=self.time_from,
            time_to=self.time_to,
            all_spikes=self.all_spikes,
            nb_neurons=self.nb_neurons,
            populations=self.populations,
            kernel=GaussianKernel(sigma=20 * ms),
        )
        isis = ISIPlot(
            (15, 2 * np.ceil(num_labelled_ct / 2)),
            scaffold=self.scaffold,
            simulation_name=self.simulation_name,
            time_from=self.time_from,
            time_to=self.time_to,
            all_spikes=self.all_spikes,
            nb_neurons=self.nb_neurons,
            populations=self.populations,
        )
        freq = FrequencyPlot(
            (15, 2 * np.ceil(num_labelled_ct / 2)),
            scaffold=self.scaffold,
            simulation_name=self.simulation_name,
            time_from=self.time_from,
            time_to=self.time_to,
            all_spikes=self.all_spikes,
            nb_neurons=self.nb_neurons,
            populations=self.populations,
        )
        legend = Legend(
            (10, 0.5 * num_labelled_ct / 3.0),
            3,
            dict_legend=dict(columnspacing=2.0, handletextpad=0.1, fontsize=20, loc="lower center"),
            dict_abbreviations=self.labelled_abbreviations,
        )
        self.add_plot("raster_psth", raster)
        self.add_plot("table", table)
        self.add_plot("firing_rates", firing_rates)
        self.add_plot("isis", isis)
        self.add_plot("freq", freq)
        self.add_plot("legend", legend)
        legend.dict_colors = raster.labelled_dict_colors.copy()
        legend.remove_ct(self.labelled_cell_names, self.ignored_ct)

    def preprocessing(self):
        self.plots["table"].set_axis_off()
        self.plots["legend"].set_axis_off()
